import re

from flask import Config
from flask_babel import lazy_gettext as _
from whoosh.analysis import CharsetFilter, LowercaseFilter, StemFilter
from whoosh.analysis.tokenizers import RegexTokenizer
from whoosh.fields import BOOLEAN, DATETIME, ID, NUMERIC, STORED, TEXT, Schema
from whoosh.query import Prefix, Term
from whoosh.support.charset import accent_map
from whoosh.util.text import rcompile

from kerko import codecs, extractors, transformers
from kerko.config_helpers import config_get
from kerko.datetime import iso_to_datetime, iso_to_timestamp
from kerko.richtext import richtext_striptags
from kerko.specs import (
    BadgeSpec,
    BibFormatSpec,
    CollectionFacetSpec,
    FacetSpec,
    FieldSpec,
    FlatFacetSpec,
    LanguageFacetSpec,
    LinkGroupSpec,
    PageSpec,
    RelationSpec,
    ScopeSpec,
    SortSpec,
    TreeFacetSpec,
)


class Composer:
    """
    A factory for the setting up the search elements.

    This class acts as a registry of search elements such as fields and facets,
    from which both the schema and the search interface can be built.

    The schema is the representation of documents and their fields in the search
    index. It is meant to be fully constructed at app configuration time and not
    changed afterwards. If schema elements need to be modified or removed, the
    application should be stopped, and the search index cleaned and rebuilt.
    """

    def __init__(self, config: Config) -> None:
        """
        Instantiate search elements based on config settings.

        This initializes elements based on config settings, but other "manually"
        instantiated elements may still be added afterwards, using the `add_*()`
        methods.
        """

        # When combining multiple strings into a single text field, the last
        # token of each string becomes adjacent to the first token of the next
        # string. This may cause phrase searches to match those tokens as if
        # they were neighbors, even though they were not in the source data. To
        # prevent this, we join the strings with the record separator character
        # and treat that character as a token. This solution is imperfect,
        # however, as the issue may still arise when a slop factor is applied to
        # the phrase search.
        token_pattern = rcompile(r"\w+(\.?\w+)*|" + re.escape(extractors.RECORD_SEPARATOR))

        # Replace the standard analyzer with one that has no stop words (helping
        # people who do phrase searches without specifying actual phrase queries).
        self.text_chain = (
            RegexTokenizer(expression=token_pattern)
            | StemFilter(lang=config_get(config, "kerko.search.whoosh_language"))
            | CharsetFilter(accent_map)
            | LowercaseFilter()
        )

        # Same for names, but without stemming.
        self.name_chain = (
            RegexTokenizer(expression=token_pattern) | CharsetFilter(accent_map) | LowercaseFilter()
        )

        self.schema = Schema()
        self.scopes: dict[str, ScopeSpec] = {}
        self.fields: dict[str, FieldSpec] = {}
        self.facets: dict[str, FacetSpec] = {}
        self.sorts: dict[str, SortSpec] = {}
        self.bib_formats: dict[str, BibFormatSpec] = {}
        self.relations: dict[str, RelationSpec] = {}
        self.badges: dict[str, BadgeSpec] = {}
        self.pages: dict[str, PageSpec] = {}
        self.link_groups: dict[str, LinkGroupSpec] = {}
        self.init_scopes(config)
        self.init_fields(config)
        self.init_facets(config)
        self.init_sorts(config)
        self.init_bib_formats(config)
        self.init_relations(config)
        self.init_pages(config)
        self.init_link_groups(config)

    def init_scopes(self, config: Config) -> None:
        """
        Initialize a set of `ScopeSpec` instances using config settings.
        """
        # Note: Default labels are defined here rather than in config so that they are translatable.
        selector_label = {
            "all": _("Everywhere"),
            "creator": _("In authors or contributors"),
            "title": _("In titles"),
            "pubyear": _("In publication years"),
            "metadata": _("In all fields"),
            "fulltext": _("In documents"),
        }
        breadbox_label = {
            "all": _("Everywhere"),
            "creator": _("In authors or contributors"),
            "title": _("In titles"),
            "pubyear": _("In publication years"),
            "metadata": _("In all fields"),
            "fulltext": _("In documents"),
        }
        help_text = {
            "all": _(
                "Search your keywords in all bibliographic record fields "
                "and in the text content of the available documents."
            ),
            "creator": _("Search your keywords in author or contributor names."),
            "title": _("Search your keywords in titles."),
            "pubyear": _(
                "Search a specific publication year (you may use the <strong>%(or_op)s</strong> "
                "operator with your keywords to find records having different publication years, "
                "e.g., <code>2020 %(or_op)s 2021</code>).",
                or_op=_("OR"),
            ),
            "metadata": _("Search your keywords in all bibliographic record fields."),
            "fulltext": _("Search your keywords in the text content of the available documents."),
        }
        scopes_dict = config_get(config, "kerko.scopes")
        for scope_key, scope_config in scopes_dict.items():
            if scope_config["enabled"]:
                kwargs = {
                    "weight": scope_config["weight"],
                }
                kwargs["selector_label"] = scope_config.get("selector_label") or selector_label.get(
                    scope_key, scope_key
                )
                kwargs["breadbox_label"] = scope_config.get("breadbox_label") or breadbox_label.get(
                    scope_key, scope_key
                )
                kwargs["help_text"] = scope_config.get("help_text") or help_text.get(scope_key, "")
                self.add_scope(ScopeSpec(key=scope_key, **kwargs))

    def init_fields(self, config: Config) -> None:
        """
        Initialize a set of `FieldSpec` instances using config settings.
        """

        #
        # Required searchable fields (partially configurable).
        #

        # Primary ID used for resolving items. Same as the Zotero item key.
        field_dict = config_get(config, "kerko.search_fields.core.required.id")
        self.add_field(
            FieldSpec(
                key="id",
                field_type=ID(unique=True, stored=True, field_boost=field_dict["boost"]),
                scopes=field_dict["scopes"],
                extractor=extractors.ItemExtractor(key="key"),
            )
        )
        # Alternate IDs used when the primary ID cannot be resolved.
        field_dict = config_get(config, "kerko.search_fields.core.required.alternate_id")
        self.add_field(
            FieldSpec(
                key="alternate_id",
                field_type=ID(field_boost=field_dict["boost"]),
                scopes=field_dict["scopes"],
                extractor=extractors.MultiExtractor(
                    extractors=[
                        extractors.ItemDataExtractor(key="DOI"),
                        extractors.ItemDataExtractor(key="ISBN"),
                        extractors.ItemDataExtractor(key="ISSN"),
                        # Extract DOI, ISBN, ISSN from the Extra field.
                        extractors.TransformerExtractor(
                            extractor=extractors.ItemDataExtractor(key="extra"),
                            transformers=[
                                transformers.find(
                                    regex=r"^\s*(DOI|ISBN|ISSN):\s*(\S+)\s*$",
                                    flags=re.IGNORECASE | re.MULTILINE,
                                    group=2,
                                    max_matches=0,
                                ),
                            ],
                        ),
                        # Extract dc:replaces relations.
                        extractors.TransformerExtractor(
                            extractor=extractors.ItemRelationsExtractor(predicate="dc:replaces"),
                            transformers=[transformers.find_item_id_in_zotero_uris_list],
                        ),
                    ]
                ),
            )
        )
        # Item type label, searchable and stored.
        field_dict = config_get(config, "kerko.search_fields.core.required.item_type_label")
        self.add_field(
            FieldSpec(
                key="item_type_label",
                field_type=TEXT(
                    analyzer=self.text_chain, stored=True, field_boost=field_dict["boost"]
                ),
                scopes=field_dict["scopes"],
                extractor=extractors.ItemTypeLabelExtractor(),
            )
        )
        # Publication year, based on a parsing of Zotero's Date field, searchable and stored.
        field_dict = config_get(config, "kerko.search_fields.core.required.year")
        self.add_field(
            FieldSpec(
                key="year",
                field_type=ID(stored=True, field_boost=field_dict["boost"]),
                scopes=field_dict["scopes"],
                extractor=extractors.YearExtractor(),
            )
        )

        #
        # Optional searchable fields (partially configurable).
        #

        field_dict = config_get(config, "kerko.search_fields.core.optional.creator")
        if field_dict["enabled"]:
            self.add_field(
                FieldSpec(
                    key="text_creator",
                    field_type=TEXT(analyzer=self.name_chain, field_boost=field_dict["boost"]),
                    scopes=field_dict["scopes"],
                    extractor=extractors.CreatorsExtractor(),
                )
            )
        field_dict = config_get(config, "kerko.search_fields.core.optional.collections")
        if field_dict["enabled"]:
            self.add_field(
                FieldSpec(
                    key="text_collections",
                    field_type=TEXT(analyzer=self.text_chain, field_boost=field_dict["boost"]),
                    scopes=field_dict["scopes"],
                    extractor=extractors.CollectionNamesExtractor(),
                )
            )
        field_dict = config_get(config, "kerko.search_fields.core.optional.tags")
        if field_dict["enabled"]:
            self.add_field(
                FieldSpec(
                    key="text_tags",
                    field_type=TEXT(analyzer=self.text_chain, field_boost=field_dict["boost"]),
                    scopes=field_dict["scopes"],
                    extractor=extractors.TagsTextExtractor(
                        include_re=config_get(config, "kerko.zotero.tag_include_re"),
                        exclude_re=config_get(config, "kerko.zotero.tag_exclude_re"),
                    ),
                )
            )
        field_dict = config_get(config, "kerko.search_fields.core.optional.notes")
        if field_dict["enabled"]:
            self.add_field(
                FieldSpec(
                    key="text_notes",
                    field_type=TEXT(analyzer=self.text_chain, field_boost=field_dict["boost"]),
                    scopes=field_dict["scopes"],
                    extractor=extractors.ChildNotesTextExtractor(
                        include_re=config_get(config, "kerko.zotero.child_include_re"),
                        exclude_re=config_get(config, "kerko.zotero.child_exclude_re"),
                    ),
                )
            )
        field_dict = config_get(config, "kerko.search_fields.core.optional.documents")
        if field_dict["enabled"]:
            self.add_field(
                FieldSpec(
                    key="text_docs",
                    field_type=TEXT(analyzer=self.text_chain, field_boost=field_dict["boost"]),
                    scopes=field_dict["scopes"],
                    extractor=extractors.ChildAttachmentsFulltextExtractor(
                        mime_types=config_get(config, "kerko.zotero.attachment_mime_types"),
                        include_re=config_get(config, "kerko.zotero.child_include_re"),
                        exclude_re=config_get(config, "kerko.zotero.child_exclude_re"),
                    ),
                )
            )

        #
        # Required relation fields, searchable for internal purposes only (hence
        # the absence of a 'scopes' parameter) (non-configurable).
        #

        # References to items that are cited by the item.
        self.add_field(
            FieldSpec(
                key="rel_cites",
                field_type=ID(stored=True),
                extractor=extractors.RelationsInChildNotesExtractor(
                    include_re=r"_cites", exclude_re=""
                ),
            )
        )
        # Items related through Zotero's relation field.
        self.add_field(
            FieldSpec(
                key="rel_related",
                field_type=ID(stored=True),
                extractor=extractors.TransformerExtractor(
                    extractor=extractors.ItemRelationsExtractor(
                        predicate="dc:relation",
                    ),
                    transformers=[transformers.find_item_id_in_zotero_uris_list],
                ),
            )
        )

        #
        # Required stored fields (unsearchable, non-configurable).
        #

        self.add_field(
            FieldSpec(
                key="item_type",
                field_type=ID(stored=True),
                extractor=extractors.ItemDataExtractor(key="itemType"),
            )
        )
        self.add_field(
            FieldSpec(
                key="date_added",
                field_type=STORED,
                extractor=extractors.ItemDataExtractor(key="dateAdded"),
            )
        )
        self.add_field(
            FieldSpec(
                key="date_modified",
                field_type=STORED,
                extractor=extractors.ItemDataExtractor(key="dateModified"),
            )
        )
        # URL from Zotero's URL field.
        self.add_field(
            FieldSpec(
                key="url",
                field_type=STORED,
                extractor=extractors.ItemDataExtractor(key="url"),
            )
        )
        # Formatted citation.
        self.add_field(
            FieldSpec(
                key="bib",
                field_type=STORED,
                extractor=extractors.ItemExtractor(key="bib", format_="bib"),
            )
        )
        # OpenURL Coins.
        self.add_field(
            FieldSpec(
                key="coins",
                field_type=STORED,
                extractor=extractors.ItemExtractor(key="coins", format_="coins"),
            )
        )
        # RIS.
        self.add_field(
            FieldSpec(
                key="ris",
                field_type=STORED,
                extractor=extractors.ItemExtractor(key="ris", format_="ris"),
            )
        )
        # BibTeX.
        self.add_field(
            FieldSpec(
                key="bibtex",
                field_type=STORED,
                extractor=extractors.ItemExtractor(key="bibtex", format_="bibtex"),
            )
        )
        # Raw item data.
        self.add_field(
            FieldSpec(
                key="data",
                field_type=STORED,
                extractor=extractors.RawDataExtractor(),
            )
        )
        # Child notes of the item.
        self.add_field(
            FieldSpec(
                key="notes",
                field_type=STORED,
                extractor=extractors.RawChildNotesExtractor(
                    include_re=config_get(config, "kerko.zotero.child_include_re"),
                    exclude_re=config_get(config, "kerko.zotero.child_exclude_re"),
                ),
            )
        )
        # URL attachments of the item.
        self.add_field(
            FieldSpec(
                key="links",
                field_type=STORED,
                extractor=extractors.ChildLinkedURIAttachmentsExtractor(
                    include_re=config_get(config, "kerko.zotero.child_include_re"),
                    exclude_re=config_get(config, "kerko.zotero.child_exclude_re"),
                ),
            )
        )
        # File attachments of the item.
        self.add_field(
            FieldSpec(
                key="attachments",
                field_type=STORED,
                extractor=extractors.ChildFileAttachmentsExtractor(
                    mime_types=config_get(config, "kerko.zotero.attachment_mime_types"),
                    include_re=config_get(config, "kerko.zotero.child_include_re"),
                    exclude_re=config_get(config, "kerko.zotero.child_exclude_re"),
                ),
            )
        )
        # Fields and labels for this item type, for convenient access.
        self.add_field(
            FieldSpec(
                key="item_fields",
                field_type=STORED,
                extractor=extractors.ItemFieldsExtractor(),
            )
        )
        # Creator types for this item type, for convenient access.
        self.add_field(
            FieldSpec(
                key="creator_types",
                field_type=STORED,
                extractor=extractors.CreatorTypesExtractor(),
            )
        )
        # URL for opening item in Zotero app.
        self.add_field(
            FieldSpec(
                key="zotero_app_url",
                field_type=STORED,
                extractor=extractors.ZoteroAppItemURLExtractor(),
            )
        )
        # URL of the item on zotero.org.
        self.add_field(
            FieldSpec(
                key="zotero_web_url",
                field_type=STORED,
                extractor=extractors.ZoteroWebItemURLExtractor(),
            )
        )

        #
        # Optional searchable fields from Zotero items (configurable).
        #

        # Those field names are prefixed with 'z_' in the search schema to
        # prevent clashes with other fields should Zotero's schema change.
        zotero_fields_dict = config_get(config, "kerko.search_fields.zotero")
        for field_key, field_config in zotero_fields_dict.items():
            if field_config["enabled"]:
                analyzer = field_config["analyzer"]
                if analyzer == "id":
                    # Identifier fields are indexed as-is.
                    self.add_field(
                        FieldSpec(
                            key=f"z_{field_key}",
                            field_type=ID(field_boost=field_config["boost"]),
                            scopes=field_config["scopes"],
                            extractor=extractors.ItemDataExtractor(key=field_key),
                        )
                    )
                elif analyzer == "text":
                    # Text fields go through the text tokenization, stemming, etc.
                    self.add_field(
                        FieldSpec(
                            key=f"z_{field_key}",
                            field_type=TEXT(
                                analyzer=self.text_chain,
                                field_boost=field_config["boost"],
                            ),
                            scopes=field_config["scopes"],
                            extractor=extractors.TransformerExtractor(
                                extractor=extractors.ItemDataExtractor(key=field_key),
                                transformers=[richtext_striptags],
                            ),
                        )
                    )
                elif analyzer == "name":
                    # Name fields are handled like text, but without stemming.
                    self.add_field(
                        FieldSpec(
                            key=f"z_{field_key}",
                            field_type=TEXT(
                                analyzer=self.name_chain,
                                field_boost=field_config["boost"],
                            ),
                            scopes=field_config["scopes"],
                            extractor=extractors.TransformerExtractor(
                                extractor=extractors.ItemDataExtractor(key=field_key),
                                transformers=[richtext_striptags],
                            ),
                        )
                    )

        #
        # Required fields for sorting (non-configurable).
        #

        self.add_field(
            FieldSpec(
                key="sort_title",
                field_type=TEXT(phrase=False, sortable=True),
                extractor=extractors.SortTitleExtractor(),
            )
        )
        self.add_field(
            FieldSpec(
                key="sort_creator",
                field_type=TEXT(phrase=False, sortable=True),
                extractor=extractors.SortCreatorExtractor(),
            )
        )
        self.add_field(
            FieldSpec(
                key="sort_date",
                field_type=NUMERIC(sortable=True),
                extractor=extractors.SortDateExtractor(),
            )
        )
        self.add_field(
            FieldSpec(
                key="sort_date_added",
                field_type=NUMERIC(sortable=True),
                extractor=extractors.TransformerExtractor(
                    extractor=extractors.ItemDataExtractor(key="dateAdded"),
                    transformers=[iso_to_timestamp],
                ),
            )
        )
        self.add_field(
            FieldSpec(
                key="sort_date_modified",
                field_type=NUMERIC(sortable=True),
                extractor=extractors.TransformerExtractor(
                    extractor=extractors.ItemDataExtractor(key="dateModified"),
                    transformers=[iso_to_timestamp],
                ),
            )
        )

        #
        # Required fields for internal filtering (non-configurable).
        #

        self.add_field(
            FieldSpec(
                key="filter_date",
                field_type=DATETIME,
                extractor=extractors.ChainExtractor(
                    extractors=[
                        extractors.MaximizeParsedDateExtractor(),
                        extractors.TransformerExtractor(
                            extractor=extractors.ItemDataExtractor(key="dateAdded"),
                            transformers=[iso_to_datetime],
                        ),
                    ]
                ),
            )
        )

    def init_facets(self, config: Config) -> None:
        """
        Initialize a set of `FacetSpec` instances using config settings.
        """
        # Note: Default titles are defined here rather than in config so that they are translatable.
        # TODO: Refactor facet using factory methods in the models, as in init_link_groups().
        facets_dict = config_get(config, "kerko.facets")
        for facet_key, facet_config in facets_dict.items():
            if facet_config["enabled"]:
                facet_type = facet_config["type"]
                kwargs = {
                    k: v for k, v in facet_config.items() if k not in ["enabled", "title", "type"]
                }
                if facet_type == "tag":
                    self.add_facet(
                        FlatFacetSpec(
                            key=f"facet_{facet_key}",
                            field_type=ID(stored=True),
                            extractor=extractors.TagsFacetExtractor(
                                include_re=config_get(config, "kerko.zotero.tag_include_re"),
                                exclude_re=config_get(config, "kerko.zotero.tag_exclude_re"),
                            ),
                            codec=codecs.BaseFacetCodec(),
                            title=facet_config.get("title") or _("Topic"),
                            missing_label=None,  # TODO:config: Allow in config.
                            allow_overlap=True,
                            query_class=Term,
                            **kwargs,
                        )
                    )
                elif facet_type == "item_type":
                    self.add_facet(
                        FlatFacetSpec(
                            key=f"facet_{facet_key}",
                            field_type=ID(stored=True),
                            extractor=extractors.ItemTypeFacetExtractor(),
                            codec=codecs.ItemTypeFacetCodec(),
                            title=facet_config.get("title") or _("Resource type"),
                            missing_label=None,  # TODO:config: Allow in config.
                            allow_overlap=False,
                            query_class=Prefix,
                            **kwargs,
                        )
                    )
                elif facet_type == "year":
                    self.add_facet(
                        TreeFacetSpec(
                            key=f"facet_{facet_key}",
                            field_type=ID(stored=True),
                            extractor=extractors.YearFacetExtractor(),
                            codec=codecs.YearTreeFacetCodec(),
                            title=facet_config.get("title") or _("Publication year"),
                            missing_label=_("Unknown"),  # TODO:config: Allow in config.
                            allow_overlap=True,
                            query_class=Prefix,
                            **kwargs,
                        )
                    )
                elif facet_type == "language":
                    self.add_facet(
                        LanguageFacetSpec(
                            key=f"facet_{facet_key}",
                            title=facet_config.get("title") or _("Resource language"),
                            missing_label=None,  # TODO:config: Allow in config.
                            **kwargs,
                        )
                    )
                elif facet_type == "link":
                    self.add_facet(
                        FlatFacetSpec(
                            key=f"facet_{facet_key}",
                            field_type=BOOLEAN(stored=True),
                            extractor=extractors.ItemDataLinkFacetExtractor(key="url"),
                            codec=codecs.BooleanFacetCodec(),
                            title=facet_config.get("title") or _("Online resource"),
                            missing_label=None,
                            allow_overlap=False,
                            query_class=Term,
                            **kwargs,
                        )
                    )
                elif facet_type == "collection":
                    self.add_facet(
                        CollectionFacetSpec(
                            key=f"facet_{facet_key}",
                            title=facet_config["title"],
                            missing_label=None,  # TODO:config: Allow in config.
                            **kwargs,
                        )
                    )

    def init_sorts(self, config: Config) -> None:
        """
        Initialize a set of `SortSpec` instances using config settings.

        These rely on `FieldSpec` instances, which must have been added beforehand.
        """
        # Note: Default labels are defined here rather than in config so that they are translatable.
        sorts_dict = config_get(config, "kerko.sorts")
        for sort_key, sort_config in sorts_dict.items():
            if sort_config["enabled"]:
                if sort_key == "score":
                    self.add_sort(
                        SortSpec(
                            key=sort_key,
                            label=sort_config.get("label") or _("Relevance"),
                            weight=sort_config["weight"],
                            fields=None,
                            # Sort by score is only possible on keyword search.
                            is_allowed=lambda criteria: criteria.has_keywords(),
                        )
                    )
                elif sort_key == "recently_added":
                    self.add_sort(
                        SortSpec(
                            key=sort_key,
                            label=sort_config.get("label") or _("Recently added"),
                            weight=sort_config["weight"],
                            fields=[
                                self.fields["sort_date_added"],
                                self.fields["sort_date"],
                                self.fields["sort_creator"],
                                self.fields["sort_title"],
                            ],
                            reverse=[
                                True,
                                True,
                                False,
                                False,
                            ],
                        )
                    )
                elif sort_key == "date_desc":
                    self.add_sort(
                        SortSpec(
                            key=sort_key,
                            label=sort_config.get("label") or _("Newest first"),
                            weight=sort_config["weight"],
                            fields=[
                                self.fields["sort_date"],
                                self.fields["sort_creator"],
                                self.fields["sort_title"],
                            ],
                            reverse=[
                                True,
                                False,
                                False,
                            ],
                        )
                    )
                elif sort_key == "date_asc":
                    self.add_sort(
                        SortSpec(
                            key=sort_key,
                            label=sort_config.get("label") or _("Oldest first"),
                            weight=sort_config["weight"],
                            fields=[
                                self.fields["sort_date"],
                                self.fields["sort_creator"],
                                self.fields["sort_title"],
                            ],
                        )
                    )
                elif sort_key == "author_asc":
                    self.add_sort(
                        SortSpec(
                            key=sort_key,
                            label=sort_config.get("label") or _("Author A-Z"),
                            weight=sort_config["weight"],
                            fields=[
                                self.fields["sort_creator"],
                                self.fields["sort_title"],
                                self.fields["sort_date"],
                            ],
                        )
                    )
                elif sort_key == "author_desc":
                    self.add_sort(
                        SortSpec(
                            key=sort_key,
                            label=sort_config.get("label") or _("Author Z-A"),
                            weight=sort_config["weight"],
                            fields=[
                                self.fields["sort_creator"],
                                self.fields["sort_title"],
                                self.fields["sort_date"],
                            ],
                            reverse=[
                                True,
                                False,
                                False,
                            ],
                        )
                    )
                elif sort_key == "title_asc":
                    self.add_sort(
                        SortSpec(
                            key=sort_key,
                            label=sort_config.get("label") or _("Title A-Z"),
                            weight=sort_config["weight"],
                            fields=[
                                self.fields["sort_title"],
                                self.fields["sort_creator"],
                                self.fields["sort_date"],
                            ],
                        )
                    )
                elif sort_key == "title_desc":
                    self.add_sort(
                        SortSpec(
                            key=sort_key,
                            label=sort_config.get("label") or _("Title Z-A"),
                            weight=sort_config["weight"],
                            fields=[
                                self.fields["sort_title"],
                                self.fields["sort_creator"],
                                self.fields["sort_date"],
                            ],
                            reverse=[
                                True,
                                False,
                                False,
                            ],
                        )
                    )

    def init_bib_formats(self, config: Config) -> None:
        """
        Initialize a set of `BibFormatSpec` instances using config settings.

        These rely on `FieldSpec` instances, which must have been added beforehand.
        """
        formats_dict = config_get(config, "kerko.bib_formats")
        for format_key, format_config in formats_dict.items():
            if format_config["enabled"]:
                kwargs = {
                    k: v
                    for k, v in format_config.items()
                    if k in ["weight", "extension", "mime_type"]
                }
                if format_key == "ris":
                    self.add_bib_format(
                        BibFormatSpec(
                            key=format_key,
                            field=self.fields["ris"],
                            label=format_config.get("label") or _("RIS"),
                            help_text=format_config.get("help_text")
                            or _("Recommended format for most reference management software"),
                            **kwargs,
                        )
                    )
                elif format_key == "bibtex":
                    self.add_bib_format(
                        BibFormatSpec(
                            key=format_key,
                            field=self.fields["bibtex"],
                            label=format_config.get("label") or _("BibTeX"),
                            help_text=format_config.get("help_text")
                            or _("Recommended format for BibTeX-specific software"),
                            **kwargs,
                        )
                    )

    def init_relations(self, config: Config) -> None:
        """
        Initialize a set of `RelationSpec` instances using config settings.

        These rely on `FieldSpec` instances, which must have been added beforehand.
        """
        relations_dict = config_get(config, "kerko.relations")
        for rel_key, rel_config in relations_dict.items():
            if rel_config["enabled"]:
                if rel_key == "cites":
                    self.add_relation(
                        RelationSpec(
                            key=rel_key,
                            field=self.fields["rel_cites"],
                            label=rel_config.get("label") or _("Cites"),
                            weight=rel_config["weight"],
                            id_fields=[self.fields["id"], self.fields["alternate_id"]],
                            reverse=True,
                            reverse_key="isCitedBy",
                            reverse_field_key="rev_cites",
                            reverse_label=_("Cited by"),
                        )
                    )
                elif rel_key == "related":
                    self.add_relation(
                        RelationSpec(
                            key=rel_key,
                            field=self.fields["rel_related"],
                            label=rel_config.get("label") or _("Related"),
                            weight=rel_config["weight"],
                            id_fields=[self.fields["id"]],
                            directed=False,
                        )
                    )

    def init_pages(self, config: Config) -> None:
        pages = config["kerko_config"].kerko.pages
        self.pages = pages.to_spec() if pages else {}

    def init_link_groups(self, config: Config) -> None:
        self.link_groups = config["kerko_config"].kerko.link_groups.to_spec()

    def add_scope(self, scope):
        self.scopes[scope.key] = scope

    def remove_scope(self, key):
        del self.scopes[key]

    def add_field(self, field):
        self.fields[field.key] = field
        if field.field_type:
            self.schema.add(field.key, field.field_type)

    def remove_field(self, key):
        self.schema.remove(key)
        del self.fields[key]

    def select_fields(self, keys):
        """
        Return a subset of specifications.

        :param list keys: Keys of the desired specs. Key that don't exist in the
            specifications dict are silently ignored.

        :return dict: The desired specs.
        """
        return {key: self.fields[key] for key in self.fields.keys() & keys}

    def add_facet(self, facet):
        self.facets[facet.key] = facet
        self.schema.add(facet.key, facet.field_type)

    def remove_facet(self, key):
        self.schema.remove(key)
        del self.facets[key]

    def add_sort(self, sort):
        self.sorts[sort.key] = sort

    def remove_sort(self, key):
        del self.sorts[key]

    def add_bib_format(self, bib_format):
        self.bib_formats[bib_format.key] = bib_format

    def remove_bib_format(self, key):
        del self.bib_formats[key]

    def add_badge(self, badge):
        self.badges[badge.key] = badge

    def remove_badge(self, key):
        del self.badges[key]

    def add_relation(self, relation):
        self.relations[relation.key] = relation

    def remove_relation(self, key):
        del self.relations[key]

    def add_page(self, key: str, page: PageSpec):
        self.pages[key] = page

    def remove_page(self, key: str):
        del self.pages[key]

    def add_link_group(self, key: str, link_group: LinkGroupSpec):
        self.link_groups[key] = link_group

    def remove_link_group(self, key: str):
        del self.link_groups[key]

    def get_ordered_specs(self, attr):
        """
        Return a list of specifications, sorted by weight.

        :param str attr: Attribute name of the specifications dict. The
            specifications must themselves have a `weight` attribute.
        """
        return sorted(getattr(self, attr).values(), key=lambda spec: spec.weight)



"""
Data transformation utilities.
"""

import itertools
import re


def find(regex, flags=0, group=1, max_matches=1, iterate=False):
    """
    Return a callable that finds all non-overlapping matches in a given string.

    :param str regex: Regular expression to search.

    :param int flags: Flags controlling the regular expression's behavior.

    :param int group: Number of the match group to retrieve from the match
    object.

    :param int max_matches: Maximum number of matching strings to return. A
        value of `0` means no maximum. If set to `1` and `iterate` is `False`,
        the returned value will be a string; otherwise a list.

    :param bool iterate: If `True`, search values within an iterable, and
        return a list of matches across all of the iterable's values.
    """
    regex = re.compile(regex, flags)

    def _find(value):
        matches = []
        if value and isinstance(value, str):
            for i, match in enumerate(regex.finditer(value)):
                if max_matches and i == max_matches:
                    break
                if match:
                    matches.append(match.group(group))
        if max_matches > 0:
            matches = matches[:max_matches]
        if max_matches == 1 and not iterate:
            if matches:
                return matches[0]
            return ""
        return matches

    def _iterate_find(value):
        assert not isinstance(value, str)  # Not expecting to process strings here.
        return list(itertools.chain.from_iterable(filter(None, map(_find, value))))

    if iterate:
        return _iterate_find
    return _find


def split(sep=None, maxsplit=-1):
    def _split(value):
        if isinstance(value, str):
            return [v.strip() for v in value.split(sep, maxsplit)]
        return []

    return _split


ZOTERO_URI_REGEX = (
    r"(https?://(www\.)?zotero\.org/|zotero://select/)"
    r"(library|((groups|users)/[0-9]+))/items/([A-Z0-9]+)"
)
ZOTERO_URI_REGEX_GROUP_ITEM_ID = 6

find_item_id_in_zotero_uris_list = find(
    regex=r"(?:^|\s)" + ZOTERO_URI_REGEX + r"(?=$|\s)",
    flags=re.MULTILINE,
    group=ZOTERO_URI_REGEX_GROUP_ITEM_ID,
    max_matches=1,
    iterate=True,
)

find_item_id_in_zotero_uris_str = find(
    regex=r"(?:^|\s)" + ZOTERO_URI_REGEX + r"(?=$|\s)",
    flags=re.MULTILINE,
    group=ZOTERO_URI_REGEX_GROUP_ITEM_ID,
    max_matches=0,
)

find_item_id_in_zotero_uri_links = find(
    regex=r'<a\s+(?:[^>]*?\s+)?href=["\']' + ZOTERO_URI_REGEX + r'["\']',
    group=ZOTERO_URI_REGEX_GROUP_ITEM_ID,
    max_matches=0,
)



from functools import partial

from flask import Blueprint as BaseBlueprint
from flask import Config
from flask.app import App

from kerko import jinja2
from kerko.views.routes import page
from kerko.views.urls import urls


class Blueprint(BaseBlueprint):
    def register(self, app: App, options: dict) -> None:
        jinja2.register_filters(self)
        jinja2.register_globals(self)
        self._add_core_urls()
        self._add_page_urls(app.config)
        super().register(app, options)

    def _add_core_urls(self) -> None:
        """
        Register Kerko's URLs.

        This must be called before the blueprint is registered on the app.
        """
        for rule_kwargs in urls:
            self.add_url_rule(**rule_kwargs)

    def _add_page_urls(self, config: Config) -> None:
        """
        Register pages defined in configuration.

        This must be called before the blueprint is registered on the app.
        """
        if "kerko_composer" in config:
            for key, page_spec in config["kerko_composer"].pages.items():
                self.add_url_rule(
                    rule=page_spec.path,
                    endpoint=key,
                    view_func=partial(page, item_id=page_spec.item_id, title=page_spec.title),
                )



"""Rich text utilities."""

import re
from typing import Any

from markupsafe import Markup

_richtext_base_tags_unescaped = re.compile(r"<(/*)\s*(i|b|sub|sup)\s*>")
_richtext_base_tags_escaped = re.compile(r"&lt;(/*)\s*(i|b|sub|sup)\s*&gt;")

_richtext_span_nocase_unescaped = re.compile(
    r'<span\s+class\s*=\s*"\s*nocase\s*"\s*>(.*?)</\s*span\s*>',
    re.DOTALL,
)
_richtext_span_nocase_escaped = re.compile(
    r"&lt;span\s+class\s*=\s*&#34;\s*nocase\s*&#34;\s*&gt;(.*?)&lt;/\s*span\s*&gt;",
    re.DOTALL,
)

_richtext_span_smallcaps_unescaped = re.compile(
    r'<span\s+style\s*=\s*"\s*font-variant\s*:small-caps;\s*"\s*>(.*?)</\s*span\s*>',
    re.DOTALL,
)
_richtext_span_smallcaps_escaped = re.compile(
    r"&lt;span\s+style\s*=\s*&#34;\s*font-variant\s*:small-caps;\s*&#34;\s*&gt;(.*?)&lt;/\s*span\s*&gt;",
    re.DOTALL,
)


def richtext_escape(s: Any) -> Markup:
    """
    Escape HTML markup, except for rich text formatting elements.

    Escape characters ``&``, ``<``, ``>``, ``'``, and ``"`` in the input, but
    allow unescaped ``<i>``, ``<b>``, ``<sub>``, ``<sup>``, ``<span
    style="font-variant:small-caps;">``, and ``<span class="nocase">`` elements,
    which are allowed by Zotero for rich text formatting.

    Reference: https://www.zotero.org/support/kb/rich_text_bibliography.

    This function's interface is modeled after that of ``Markup.escape()``.
    """
    if hasattr(s, "__html__"):
        return Markup(s.__html__())  # s is already marked as safe.

    # It is hard to selectively escape elements. Instead, this escapes them all
    # first, and then un-escapes those that match specific regular expressions.
    s = (
        str(s)
        .replace("&", "&amp;")
        .replace(">", "&gt;")
        .replace("<", "&lt;")
        .replace("'", "&#39;")
        .replace('"', "&#34;")
    )
    s = _richtext_base_tags_escaped.sub(r"<\1\2>", s)
    s = _richtext_span_nocase_escaped.sub(r'<span class="nocase">\1</span>', s)
    s = _richtext_span_smallcaps_escaped.sub(r'<span style="font-variant:small-caps;">\1</span>', s)
    return Markup(s)


def richtext_striptags(s: str) -> str:
    """
    Strip rich text formatting elements from the given string.

    HTML elements that are NOT useful to rich text formatting are left alone and
    are considered as literal text.

    Reference: https://www.zotero.org/support/kb/rich_text_bibliography.

    Unlike ``Markup.striptags()``, this function does not assume that the input
    is HTML markup, and it does not unescape HTML entities.
    """
    s = _richtext_base_tags_unescaped.sub("", s)
    s = _richtext_span_nocase_unescaped.sub(r"\1", s)
    s = _richtext_span_smallcaps_unescaped.sub(r"\1", s)
    return s



import re
from abc import ABC, abstractmethod

from flask import render_template, render_template_string


class Renderer(ABC):
    @abstractmethod
    def render(self, **context):
        pass


class TemplateRenderer:
    """Render using a template file."""

    def __init__(self, template, **extra_context):
        self.template = template
        self.extra_context = extra_context

    def render(self, **context):
        return render_template(self.template, **context, **self.extra_context)


class TemplateResolverRenderer:
    """Render using a template file whose name is resolved from context."""

    def __init__(self, template, **extra_context):
        """
        Initialize the object.

        :param str template: A template name with replacement fields delimited
            with braces (i.e. à la `str.format`). Those will be replaced at
            rendering time with values from context to determine the template
            name.
        """
        self.template = template
        self.extra_context = extra_context
        self.fields = re.findall(r"\{(.*?)\}", self.template)

    def render(self, **context):
        return render_template(
            self.template.format(**{kw: context[kw] for kw in self.fields if kw in context}),
            **context,
            **self.extra_context,
        )


class TemplateStringRenderer:
    """Render using a template string."""

    def __init__(self, template_string, **extra_context):
        self.template_string = template_string
        self.extra_context = extra_context

    def render(self, **context):
        return render_template_string(self.template_string, **context, **self.extra_context)



from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any, Optional

from babel.numbers import format_decimal
from flask import Request, url_for
from flask_babel import get_locale
from w3lib.url import safe_url_string
from werkzeug.datastructures import MultiDict
from whoosh.fields import ID
from whoosh.query import Prefix, Term

from kerko import codecs, extractors, renderers
from kerko.text import slugify, sort_normalize
from kerko.tree import Tree


class ScopeSpec:
    """
    Specifies a scope for keyword searches.

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(self, key, selector_label, breadbox_label, weight=0, help_text=""):
        self.key = key
        self.selector_label = selector_label
        self.breadbox_label = breadbox_label
        self.weight = weight
        self.help_text = help_text

    def add_keywords(self, value, active_keywords=None):
        """
        Add a value for this scope to active keywords.

        :param string value: The value to add.

        :param MultiDict active_keywords: The active keywords to derive from.

        :return MultiDict: A copy of the keywords with the added value.
        """
        new_keywords = active_keywords.deepcopy() if active_keywords else MultiDict()
        new_keywords.add(self.key, value.strip())
        return new_keywords

    def remove_keywords(self, value, active_keywords):
        """
        Remove a value for this facet from active keywords.

        :param string value: The value to remove.

        :param MultiDict active_keywords: The active keywords to derive from.

        :return MultiDict: A copy of the keywords with the value removed.
        """
        new_keywords = active_keywords.deepcopy()
        if new_keywords:
            new_values = [v for v in new_keywords.poplist(self.key) if v != value]
            if new_values:
                new_keywords.setlist(self.key, new_values)
        return new_keywords


class BaseFieldSpec(ABC):
    def __init__(
        self,
        key,
        field_type,
        extractor,
    ):
        """
        Initialize this field specification.

        :param str key: Unique key for referencing the field in the search index
            schema.

        :param whoosh.fields.FieldType field_type: Instance of the schema field
            type. Set to `None` if the field's value is to be passed along with
            documents, but not added to the schema, e.g., boost factors (see
            `whoosh.writing.IndexWriter.add_document`).

        :param kerko.extractors.Extractor extractor: Instance of the extractor
            that will extract the field value from a Zotero item.
        """
        self.key = key
        self.field_type = field_type
        self.extractor = extractor

    def extract_to_document(self, document, item, library_context):
        """Extract the value of this element from a Zotero item."""
        self.extractor.extract_and_store(document, item, library_context, self)


class FieldSpec(BaseFieldSpec):
    """Specifies a schema field."""

    def __init__(self, codec=None, scopes=None, **kwargs):
        """
        Initialize this field specification.

        :param BaseFieldCodec codec: Value encoder/decoder for this field.

        :param list scopes: list of keys to `ScopeSpec` instances to which this
            field applies. If `None`, the field won't be available for keyword
            search.
        """
        super().__init__(**kwargs)
        self.scopes = scopes or []
        self.codec = codec or codecs.IdentityFieldCodec()

    def encode(self, value):
        return self.codec.encode(value)

    def decode(self, encoded_value):
        return self.codec.decode(encoded_value)


class FacetSpec(BaseFieldSpec):
    """Specifies a facet for search grouping and filtering."""

    def __init__(
        self,
        *,
        title,
        filter_key,
        weight=0,
        initial_limit=0,
        initial_limit_leeway=2,
        codec=None,
        missing_label=None,
        sort_by=None,
        sort_reverse=False,
        item_view=True,
        allow_overlap=True,
        query_class=None,
        renderer=None,
        **kwargs,
    ):
        """
        Initialize this facet specification.

        :param str title: Title of the facet.

        :param str filter_key: Key to use in URLs when filtering with this
            facet.

        :param int weight: Determine the position of this facet relative to the
            others.

        :param int initial_limit: Maximum number of filters to show by default
            under this facet. Excess filters will be shown if the user clicks a
            "view more" button. Defaults to 0 (no limit).

        :param int initial_limit_leeway: If the number of filters under this
            facet exceeds `initial_limit` by this tolerance margin or less, all
            filters will be shown. Defaults to 0 (no tolerance margin).

        :param str missing_label: Label to use for items that do not have any
            value for this facet. Defaults to `None` (show no label at all).

        :param codecs.BaseFacetCodec codec: Value encoder/decoder for this facet.

        :param bool item_view: Show this facet on item view pages.

        :param `renderers.Renderer` renderer: A renderer for this facet. The
            rendering context provides the following variables:

            - `spec`: The `FacetSpec` instance.
            - `items`: The facet values retrieved by the search query.
            - `mode`: A string whose value is one of the following:
                - `'search'`: Displaying the facet on a search results page.
                - `'field'`: Displaying the facet on a full bibliographic record
                  page.
                - `'breadbox'`: Displaying the facet as a search criteria in the
                  breadbox.

        .. seealso: Additional :meth:`BaseFieldSpec.__init__` arguments.
        """
        super().__init__(**kwargs)
        self.title = title
        self.filter_key = filter_key
        self.weight = weight
        self.initial_limit = initial_limit
        self.initial_limit_leeway = initial_limit_leeway
        self.codec = codec or codecs.BaseFacetCodec()
        self.missing_label = missing_label
        self.sort_by = sort_by
        self.sort_reverse = sort_reverse
        self.item_view = item_view
        self.allow_overlap = allow_overlap
        self.query_class = query_class or Term
        self.renderer = renderer or renderers.TemplateResolverRenderer(
            "kerko/_facet_{mode}.html.jinja2"
        )

    def encode(self, value):
        return self.codec.encode(value)

    def decode(self, encoded_value, default_value=None, default_label=None):
        if encoded_value is None:
            return "", ""
        return self.codec.decode(encoded_value, default_value, default_label)

    @abstractmethod
    def add_filter(self, value, active_filters):
        """
        Add a value for this facet to active filters.

        :param string value: The value to add.

        :param MultiDict active_filters: The active filters to derive from.

        :return MultiDict: A copy of the filters with the added value.
        """

    @abstractmethod
    def remove_filter(self, value, active_filters):
        """
        Remove a value for this facet from active filters.

        :param string value: The value to remove.

        :param MultiDict active_filters: The active filters to derive from.

        :return MultiDict: A copy of the filters with the value removed.
        """

    @abstractmethod
    def build(self, results, criteria, active_only=False):
        """
        Construct a facet's items for display.

        :param iterable results: Iterable of (value, count) tuples representing
            the facet's results from the search query.

        :param Criteria criteria: The current search criteria. If None, the
            facet will be built for performing a new search.

        :param active_only: Only build the items that are related to active
            filters, i.e., filters actually present in the search criteria.
        """

    def sort_items(self, items):
        if self.sort_by is None:
            return items

        # Sort items based on multiple-keys.
        return sorted(
            items,
            key=lambda x: (
                # First sort key: show active items first.
                not x["remove_url"],
                # Second sort key: show items with missing labels last.
                bool(x["label"]) if self.sort_reverse else not x["label"],
                # Third sort key: sort items according to the facet's specified
                # sort keys. If 'count' is used as key, multiply the count value
                # by -1 to reverse order (because the desired default when
                # ordering by count is the descending order).
                *[
                    x[k] * -1
                    if k == "count"
                    else (sort_normalize(x[k]) if isinstance(x[k], str) else x[k])
                    for k in self.sort_by
                ],
            ),
            reverse=self.sort_reverse,
        )

    def render(self, items, mode):
        return self.renderer.render(spec=self, items=items, mode=mode)


class FlatFacetSpec(FacetSpec):
    def add_filter(self, value, active_filters):
        if value is None:  # Special case for missing value (None is returned by Whoosh).
            value = ""
        filters = active_filters.deepcopy()
        active_values = filters.getlist(self.filter_key)
        if active_values and value in active_values:
            return None  # Already filtering with value. No add needed.
        filters.setlistdefault(self.filter_key).append(value)
        return filters

    def remove_filter(self, value, active_filters):
        if value is None:  # Special case for missing value (None is returned by Whoosh).
            value = ""
        filters = active_filters.deepcopy()
        active_values = filters.getlist(self.filter_key)
        if not active_values or value not in active_values:
            return None  # Not currently filtering with value. No remove needed.
        active_values.remove(value)
        filters.setlist(self.filter_key, active_values)
        return filters

    def build(self, results, criteria, active_only=False):
        items = []
        for value, count in results.items():
            if value or self.missing_label:
                value, label = self.decode(value, default_value=value, default_label=value)  # noqa: PLW2901
                new_filters = self.remove_filter(value, criteria.filters)
                if new_filters:
                    remove_url = url_for(
                        ".search",
                        **criteria.params(
                            filters=new_filters,
                            options={
                                "page": None,
                                "page-len": None,
                                "id": None,
                            },
                        ),
                    )
                else:
                    remove_url = None
                if remove_url or active_only:
                    add_url = None
                else:
                    new_filters = self.add_filter(value, criteria.filters)
                    if new_filters:
                        add_url = url_for(
                            ".search",
                            **criteria.params(
                                filters=new_filters,
                                options={
                                    "page": None,
                                    "page-len": None,
                                    "id": None,
                                },
                            ),
                        )
                    else:
                        add_url = None
                if remove_url or add_url:  # Only items with an URL get displayed.
                    items.append(
                        {
                            "label": label,
                            "count": count,
                            "count_formatted": format_decimal(count, locale=get_locale()),
                            "remove_url": remove_url,
                            "add_url": add_url,
                        }
                    )
        return self.sort_items(items)


class TreeFacetSpec(FacetSpec):
    def __init__(self, path_separator=".", **kwargs):
        super().__init__(**kwargs)
        self.path_separator = path_separator

    @staticmethod
    def is_ancestor(ancestor, descendant):
        """
        Return True if `ancestor` is an ancestor of `descendant`.

        :param str ancestor: Potential ancestor.

        :param str descendant: Path to look for potential ancestor.
        """
        # True if ancestor is a prefix of descendant. Warning: This simple
        # condition only works for finding ancestors when all path components
        # have the same length, otherwise any partial prefix could match.
        return descendant.find(ancestor) == 0 and ancestor != descendant

    def get_parent(self, value):
        parent = value.rsplit(sep=self.path_separator, maxsplit=1)[0]
        return parent if parent != value else None

    def add_filter(self, value, active_filters):
        if value is None:
            # Special case for missing value (None is returned by Whoosh).
            value = ""
        filters = active_filters.deepcopy()
        active_values = filters.getlist(self.filter_key)
        for i, active_value in enumerate(active_values):
            if value == active_value or self.is_ancestor(value, active_value):
                # Already filtering with value or a descendant. No add needed.
                return None
            elif self.is_ancestor(active_value, value):
                # Active value is ancestor of value. Replace the active value.
                active_values[i] = value
                filters.setlist(self.filter_key, active_values)
                break
        else:
            # This is an all new filter. Add its value.
            filters.setlistdefault(self.filter_key).append(value)
        return filters

    def remove_filter(self, value, active_filters):
        if value is None:
            # Special case for missing value (None is returned by Whoosh).
            value = ""
        filters = active_filters.deepcopy()
        active_values = filters.getlist(self.filter_key)
        if not active_values:
            # Filter not active at all. No remove needed.
            return None
        new_values = []
        change = False
        for active_value in active_values:
            if value == active_value or self.is_ancestor(value, active_value):
                # Currently filtering with value or a descendant. Remove needed.
                change = True
                parent = self.get_parent(value)
                if parent and parent not in new_values:
                    new_values.append(parent)
            else:
                # Filter unrelated to value. Preserve it.
                new_values.append(active_value)
        if not change:
            # Neither filtering with value or a descendant. No remove needed.
            return None
        filters.setlist(self.filter_key, new_values)
        return filters

    def sort_tree(self, tree):
        """
        Convert tree to list, sorting children along the way.

        In the process, each item remains a dict as in a flat facet, but with
        an extra 'children' key.
        """
        lst = []
        for child in tree["children"].values():
            child["node"]["children"] = self.sort_tree(child)
            lst.append(child["node"])
        return self.sort_items(lst)

    def build(self, results, criteria, active_only=False):
        tree = Tree()
        for value, count in results.items():
            if value or self.missing_label:
                value, label = self.decode(value, default_value=value, default_label=value)  # noqa: PLW2901
                new_filters = self.remove_filter(value, criteria.filters)
                if new_filters:
                    remove_url = url_for(
                        ".search",
                        **criteria.params(
                            filters=new_filters,
                            options={
                                "page": None,
                                "page-len": None,
                                "id": None,
                            },
                        ),
                    )
                else:
                    remove_url = None
                if remove_url or active_only:
                    add_url = None
                else:
                    new_filters = self.add_filter(value, criteria.filters)
                    if new_filters:
                        add_url = url_for(
                            ".search",
                            **criteria.params(
                                filters=new_filters,
                                options={
                                    "page": None,
                                    "page-len": None,
                                    "id": None,
                                },
                            ),
                        )
                    else:
                        add_url = None
                if remove_url or add_url:  # Only items with an URL get displayed.
                    path = value.split(sep=self.path_separator)

                    # Build the tree path. Part of the path may or may not already
                    # exist as the facet values are not ordered.
                    node = tree  # Start at tree root.
                    for component in path:
                        node = node["children"][component]
                    # Add data at the leaf.
                    node["node"] = {
                        "id": "-".join(path),
                        "label": label,
                        "count": count,
                        "count_formatted": format_decimal(count, locale=get_locale()),
                        "remove_url": remove_url,
                        "add_url": add_url,
                    }
        return self.sort_tree(tree)


class LanguageFacetSpec(FlatFacetSpec):
    """
    Specifies a facet based on the Zotero language field.
    """

    def __init__(
        self, *, values_separator_re=";", normalize=True, locale="en", allow_invalid=True, **kwargs
    ):
        # Provide some convenient defaults for this type of facet.
        kwargs.setdefault("field_type", ID(stored=True))
        kwargs.setdefault("filter_key", "lang")
        kwargs.setdefault(
            "extractor",
            extractors.LanguageExtractor(
                values_separator_re=values_separator_re,
                normalize=normalize,
                locale=locale,
                allow_invalid=allow_invalid,
            ),
        )
        kwargs.setdefault("codec", codecs.LabelFacetCodec(label_separator=" "))
        kwargs.setdefault("allow_overlap", True)
        kwargs.setdefault("query_class", Prefix)
        super().__init__(**kwargs)


class CollectionFacetSpec(TreeFacetSpec):
    """
    Specifies a facet based on a top-level Zotero collection.

    A top-level Zotero collection can act as a facet when its key matches a
    given `collection_key`. Subcollections become values within the facet.
    """

    def __init__(self, *, collection_key, **kwargs):
        # Provide some convenient defaults for this type of facet.
        kwargs.setdefault("key", f"facet_collection_{collection_key}")
        kwargs.setdefault("field_type", ID(stored=True))
        kwargs.setdefault("filter_key", slugify(str(kwargs.get("title"))))
        kwargs.setdefault("codec", codecs.CollectionFacetCodec())
        kwargs.setdefault("query_class", Prefix)
        kwargs.setdefault("extractor", extractors.CollectionFacetTreeExtractor())
        super().__init__(**kwargs)
        self.collection_key = collection_key


class SortSpec:
    """
    Specifies a sort option.

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(
        self,
        key,
        label,
        fields,
        weight=0,
        reverse=False,
        is_allowed=True,
    ):
        """
        Initialize a sort option.

        :param str key: Key of this sort option.

        :param str label: Label of this sort option.

        :param list fields: list of `FieldSpec` instances to use when doing
            search queries with this sort option, in order of precedence.

        :param int weight: Determine the position of this option relative to the
            other options.

        :param bool reverse: Whether the fields should be sorted in reverse
            order. To provide per-field reverse settings, an iterable may be
            supplied instead of a bool, in which case it should contain the same
            number of elements as the `fields` parameter.

        :param callable is_allowed: Optional callable to determine if, given a
            criteria object, the sort option should be allowed.
        """
        self.key = key
        self.label = label
        self.fields = fields
        self.weight = weight
        if isinstance(reverse, Iterable):
            if all(reverse):
                self.reverse = True
            elif not any(reverse):
                self.reverse = False
            else:
                self.reverse = reverse
        else:
            self.reverse = reverse
        self._is_allowed = is_allowed

    def is_allowed(self, criteria):
        if callable(self._is_allowed):
            return self._is_allowed(criteria)
        return self._is_allowed

    def get_field_keys(self):
        if self.fields:
            return [spec.key for spec in self.fields]
        return None


class BibFormatSpec:
    """
    Specifies a bibliographic record download format.

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(
        self,
        key,
        field,
        label,
        help_text,
        weight,
        extension,
        mime_type,
        group_format="{}",
        group_item_delimiter="",
    ):
        """
        Initialize a format.

        :param str key: Key of this format.

        :param FieldSpec field: `FieldSpec` instance associated to this format.

        :param str label: Label of this format.

        :param int weight: Determine the position of this format relative to the
            others in lists.

        :param str extension: File extension of this format.

        :param str mime_type: MIME type of this format.

        :param str group_format: Format string for wrapping multiple entries.

        :param str group_item_delimiter: Delimiter string to insert between
            entries.
        """
        self.key = key
        self.field = field
        self.label = label
        self.help_text = help_text
        self.weight = weight
        self.extension = extension
        self.mime_type = mime_type
        self.group_format = group_format
        self.group_item_delimiter = group_item_delimiter


class RelationSpec:
    """
    Specifies a type of relation between items.

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(
        self,
        *,
        key,
        field,
        label,
        weight,
        id_fields,
        directed=True,
        reverse=False,
        reverse_key="",
        reverse_field_key="",
        reverse_label="",
    ):
        """
        Initialize a relation type.

        :param str key: Key of this relation type.

        :param FieldSpec field: `FieldSpec` instance associated to this relation
            type. The field will be looked up in the search index to retrieve
            item identifiers related to a given item.

        :param str label: Label of this relation type.

        :param int weight: Determine the position of this relation type
            relatively to other relation types in lists.

        :paras list id_fields: list of `FieldSpec` instances representing the
            fields to search when trying to resolve an item identifier.

        :param bool directed: Whether a relation is directed (i.e.
            bidirectional) or not.

        :param bool reverse: Whether a reverse relation should be exposed. If
            `directed` is `False`, this can only be `False` as well.

        :param str reverse_key: Key of the reverse relation. Should be set only
            if `reverse` is `True`.

        :param str reverse_field_key: Field key to use for storing the reverse
            relation. This isn't a `FieldSpec` as the field won't be looked up
            in the search index. Instead, it will be dynamically populated with
            items whose `field` contain a given item. Should be set only if
            `reverse` is `True`.

        :param str reverse_label: Label of the reverse relation. Should be set
            only if `reverse` is `True`.
        """
        assert not reverse and not directed or directed
        assert not reverse and not reverse_key or reverse
        assert not reverse and not reverse_field_key or reverse
        assert not reverse and not reverse_label or reverse
        self.key = key
        self.field = field
        self.label = label
        self.weight = weight
        self.id_fields = id_fields
        self.directed = directed
        self.reverse = reverse
        self.reverse_key = reverse_key
        self.reverse_field_key = reverse_field_key
        self.reverse_label = reverse_label


class BadgeSpec:
    """
    Specifies a badge.

    Badges may be displayed on items (in search results and on full
    bibliographic record pages).

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(
        self,
        key,
        field,
        activator,
        renderer,
        weight=0,
    ):
        """
        Initialize this badge specification.

        :param str key: Key of this badge.

        :param FieldSpec field: `FieldSpec` instance required by this badge.

        :param callable activator: Callable which, given a `FieldSpec` instance
            and an item, must return `True` if the badge should be displayed.

        :param `renderers.Renderer` renderer: A renderer for this badge. The
            rendering context provides the following variables:
            - `field`: The `FieldSpec` instance.
            - `item`: The item retrieved from the search index.
            - `mode`: A string whose value is one of the following:
                - `'result':` The item is being viewed in a list of results.
                - `'item'`: Viewing the item's full bibliographic record.

        :param int weight: Determine the position of this badge relative to the
            others.
        """
        self.key = key
        self.field = field
        self.activator = activator
        self.renderer = renderer
        self.weight = weight

    def is_active(self, item):
        """
        Return `True` is this badge is active for the given item.
        """
        return self.activator(self.field, item)

    def render(self, item, mode):
        """
        Render the badge, if necessary, for the given item.

        :return str: The rendered badge, or `''` if the badge is not activated
            on the item.
        """
        if self.is_active(item):
            return self.renderer.render(field=self.field, item=item, mode=mode)
        return ""


class PageSpec:
    """
    Specifies a page whose content is to be extracted from a Zotero note.
    """

    def __init__(self, path: str, item_id: str, title: str):
        self.path = path
        self.item_id = item_id
        self.title = title


class LinkSpec(ABC):
    def __init__(self, *, text: str, new_window=False, weight=0):
        self.text = text
        self.new_window = new_window
        self.weight = weight

    def is_active(self, request: Request) -> bool:  # noqa: ARG002
        return False

    @property
    @abstractmethod
    def url(self) -> str:
        pass


class LinkByURLSpec(LinkSpec):
    def __init__(self, *, url: str, **kwargs):
        super().__init__(**kwargs)
        self._url = url

    @property
    def url(self) -> str:
        return safe_url_string(self._url)


class LinkByEndpointSpec(LinkSpec):
    def __init__(
        self,
        *,
        endpoint: str,
        external: bool = False,
        anchor: Optional[str] = None,
        scheme: Optional[str] = None,
        parameters: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.endpoint = endpoint
        self.external = external
        self.anchor = anchor or None  # Replace empty string with None.
        self.scheme = scheme or None  # Replace empty string with None.
        self.parameters = parameters or {}

    def is_active(self, request: Request) -> bool:
        return request.endpoint == self.endpoint

    @property
    def url(self) -> str:
        return url_for(
            self.endpoint,
            _anchor=self.anchor,
            _scheme=self.scheme,
            _external=self.external,
            **self.parameters,
        )


class PageLinkSpec(LinkSpec):
    def __init__(self, *, page: str, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = f"kerko.{page}"

    def is_active(self, request: Request) -> bool:
        return request.endpoint == self.endpoint

    @property
    def url(self) -> str:
        return url_for(self.endpoint)


class LinkGroupSpec:
    def __init__(self, key: str, links: Optional[list[LinkSpec]] = None):
        self.key = key
        self.links = links or []

    def add_item(self, item: LinkSpec):
        self.links.append(item)

    def get_ordered_links(self) -> list[LinkSpec]:
        return sorted(self.links, key=lambda spec: spec.weight)



"""
Tests for item meta tags.
"""

import unittest

import elementpath  # For XPath 2.0 selectors.
from lxml import etree

from kerko.views.item.creators import format_creator_name
from tests.integration_testing import SynchronizedTestCase


class CreatorNameTestCase(unittest.TestCase):
    """Unit test for creator name formatting."""

    def test_creator_empty(self):
        self.assertEqual(
            format_creator_name({}),
            "",
        )

    def test_creator_single_field(self):
        self.assertEqual(
            format_creator_name({"name": "Full name"}),
            "Full name",
        )

    def test_creator_single_field_empty(self):
        self.assertEqual(
            format_creator_name({"name": ""}),
            "",
        )

    def test_creator_two_fields(self):
        self.assertEqual(
            format_creator_name({"firstName": "FirstName", "lastName": "LastName"}),
            "LastName, FirstName",
        )

    def test_creator_two_fields_empty(self):
        self.assertEqual(
            format_creator_name({"firstName": "", "lastName": ""}),
            "",
        )


class ItemMetaTestCase(SynchronizedTestCase):
    """Integration tests for item meta tags."""

    def test_highwire_press_book(self):
        """Test Highwire Press tags for a 'book' item."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/2351708-TAQ6HCSL", follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ["A book with multiple authors"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ["2019"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ["2019"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ["2019"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                [
                    "Mustermann, Erika",
                    "Jacques, Pierre-Jean",
                    "Pérez, Juan",
                    "Doe, John",
                    "Roe, Richard",
                    "Smith, Jane",
                    "Kumar, Ashok",
                ],
                # Google Scholar requires that only the actual authors be included.
                # Thus, 'Jansen, Jan' (Editor) is excluded.
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ["en"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publisher"]/@content'),
                ["Example Publishing"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_isbn"]/@content'),
                ["979-8292-9-6571-8"],
            )
            unexpected_tags = [
                "citation_doi",
                "citation_issn",
                "citation_volume",
                "citation_issue",
                "citation_pdf_url",
                "citation_firstpage",
                "citation_lastpage",
                "citation_conference_title",
                "citation_journal_title",
                "citation_dissertation_institution",
                "citation_technical_report_institution",
                "citation_technical_report_number",
            ]
            self.assertEqual(
                elementpath.select(
                    tree, f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
                "Tag is unexpected for this particular item, or for 'book' items in general",
            )

    def test_highwire_press_conference_paper(self):
        """Test Highwire Press tags for a 'conferencePaper' item."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/2351708-99VVIN8V", follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ["A conference paper"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ["2021-11-10"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ["2021-11-10"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ["2021"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                ["Hoe, Harry"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ["en"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publisher"]/@content'),
                ["Example Publishing"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_firstpage"]/@content'),
                ["11"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_lastpage"]/@content'),
                ["55"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_conference_title"]/@content'),
                ["First Example Conference, 2021"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_doi"]/@content'),
                ["10.5555/2351708-99VVIN8V"],
            )
            unexpected_tags = [
                "citation_isbn",
                "citation_issn",
                "citation_volume",
                "citation_issue",
                "citation_pdf_url",
                "citation_journal_title",
                "citation_dissertation_institution",
                "citation_technical_report_institution",
                "citation_technical_report_number",
                "Tag is unexpected for this particular item, "
                "or for 'conferencePaper' items in general",
            ]
            self.assertEqual(
                elementpath.select(
                    tree, f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
            )

    def test_highwire_press_journal_article(self):
        """Test Highwire Press tags for a 'journalArticle' item."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/2351708-NW89E453", follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ["Zotero: information management software 2.0"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ["2011-06-07"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ["2011-06-07"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ["2011"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                ["Fernandez, Peter"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_volume"]/@content'),
                ["28"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_issue"]/@content'),
                ["4"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ["en"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_firstpage"]/@content'),
                ["5"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_lastpage"]/@content'),
                ["7"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_journal_title"]/@content'),
                ["Library Hi Tech News"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_doi"]/@content'),
                ["10.1108/07419051111154758"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_issn"]/@content'),
                ["0741-9058"],
            )
            self.assertTrue(
                elementpath.select(
                    tree,
                    f'/html/head/meta[@name="citation_pdf_url"][matches(@content, '
                    f'"https?://[^/]+{self.URL_PREFIX}/[A-Z0-9]+/download/[A-Z0-9]+/Fernandez'
                    f'%20-%202011%20-%20Zotero%20information%20management%20software%202.0.pdf")]',
                ),
                "Incorrect PDF URL",
            )
            unexpected_tags = [
                "citation_isbn",
                "citation_publisher",
                "citation_conference_title",
                "citation_dissertation_institution",
                "citation_technical_report_institution",
                "citation_technical_report_number",
            ]
            self.assertEqual(
                elementpath.select(
                    tree, f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
                (
                    "Tag is unexpected for this particular item, "
                    "or for 'journalArticle' items in general"
                ),
            )

    def test_highwire_press_report(self):
        """Test Highwire Press tags for a 'report' item."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/2351708-DNFRSHFB", follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ["A report with a DOI in the Extra field"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ["2019"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ["2019"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ["2019"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                ["Doe, John"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ["en"],
            )
            self.assertEqual(
                tree.xpath(
                    '/html/head/meta[@name="citation_technical_report_institution"]/@content'
                ),
                ["Example Institution"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_technical_report_number"]/@content'),
                ["10"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_doi"]/@content'),
                ["10.5555/2351708-DNFRSHFB"],
            )
            unexpected_tags = [
                "citation_isbn",
                "citation_issn",
                "citation_pdf_url",
                "citation_volume",
                "citation_issue",
                "citation_firstpage",
                "citation_lastpage",
                "citation_publisher",
                "citation_journal_title",
                "citation_conference_title",
                "citation_dissertation_institution",
            ]
            self.assertEqual(
                elementpath.select(
                    tree, f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
                "Tag is unexpected for this particular item, or for 'report' items in general",
            )

    def test_highwire_press_thesis(self):
        """Test Highwire Press tags for a 'thesis' item."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/2351708-9TCEM5BE", follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ["A thesis example"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ["2015-05-15"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ["2015-05-15"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ["2015"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                ["Hoe, Harry"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ["en"],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_dissertation_institution"]/@content'),
                ["Small University"],
            )
            unexpected_tags = [
                "citation_doi",
                "citation_isbn",
                "citation_issn",
                "citation_pdf_url",
                "citation_volume",
                "citation_issue",
                "citation_firstpage",
                "citation_lastpage",
                "citation_publisher",
                "citation_journal_title",
                "citation_conference_title",
                "citation_technical_report_institution",
                "citation_technical_report_number",
            ]
            self.assertEqual(
                elementpath.select(
                    tree, f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
                "Tag is unexpected for this particular item, or for 'thesis' items in general",
            )

    def test_highwire_press_webpage(self):
        """Test Highwire Press tags for a 'webpage' item."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/4507457-YE4WVE35", follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                elementpath.select(tree, '/html/head/meta[matches(@name, "citation_.+")]/@name'),
                [],
                "Only book, conferencePaper, journalArticle, report, or thesis "
                "items are allowed to have Highwire Press tags",
            )

    # TODO: Test item with empty title/author/date.



from flask_wtf import FlaskForm
from wtforms import SelectField, validators
from wtforms.fields import SearchField

from kerko.shortcuts import composer


class SearchForm(FlaskForm):
    scope = SelectField()
    keywords = SearchField(validators=[validators.optional(), validators.length(max=1000)])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.scope.choices = [
            (s.key, s.selector_label) for s in composer().get_ordered_specs("scopes")
        ]



import pathlib
import pickle
import shutil

import whoosh

from kerko.shortcuts import data_path


class SearchIndexError(Exception):
    pass


class SchemaError(Exception):
    pass


def get_storage_dir(storage):
    return pathlib.Path(data_path()) / storage


def load_object(storage, key, default=None):
    try:
        with pathlib.Path(get_storage_dir(storage) / f"{key}.pickle").open("rb") as f:
            return pickle.load(f)
    except OSError:
        return default


def save_object(storage, key, obj):
    get_storage_dir(storage).mkdir(parents=True, exist_ok=True)
    with pathlib.Path(get_storage_dir(storage) / f"{key}.pickle").open("wb") as f:
        pickle.dump(obj, f)


def open_index(storage, *, write=False, schema=None, auto_create=False):
    """
    Open a Whoosh search index.

    :param str storage: Name of the subdirectory containing the index.

    :param dict/callable schema: Schema of the index, required if parameter
        `write` is `True`.

    :param bool auto_create: If this and `write` are `True`, create the index if
        it does not already exists.

    :param bool write: If `True`, make the index writable instead of read-only.
    """
    index_dir = get_storage_dir(storage) / "whoosh"
    try:
        index = None
        if not index_dir.exists() and auto_create and write:
            index_dir.mkdir(parents=True, exist_ok=True)
            index = whoosh.index.create_in(str(index_dir), schema() if callable(schema) else schema)
        elif index_dir.exists():
            index = whoosh.index.open_dir(str(index_dir), readonly=not write)

        if index:
            if write:  # In write mode, no further checks needed.
                return index
            if index.doc_count() > 0:  # In read mode, we expect some docs to be available.
                return index
            msg = f"Empty {storage} in directory '{index_dir}'."
            raise SearchIndexError(msg)
        raise SearchIndexError(
            f"Could not open {storage} from directory '{index_dir}'. "
            + (f"Please sync {storage}." if not write else "")
        )
    except whoosh.index.IndexError as e:
        msg = f"Error with {storage} in directory '{index_dir}': '{e}'."
        raise SearchIndexError(msg) from e


def delete_storage(storage):
    if (get_storage_dir(storage)).is_dir():
        shutil.rmtree(get_storage_dir(storage))


def get_doc_count(storage):
    """Return the number of documents in the specified Whoosh search index."""
    with open_index(storage).searcher() as searcher:
        return searcher.doc_count()



"""
Kerko: A Flask blueprint that provides faceted search for bibliographies based on Zotero.
"""

import errno
import pathlib
import sys

from kerko.blueprint import Blueprint
from kerko.config_helpers import load_toml

# Kerko won't load translations on its own. To load them, an application may add
# the following domain and translation directory to its Babel configuration.
TRANSLATION_DOMAIN = "kerko"
TRANSLATION_DIRECTORY = str(pathlib.Path(__file__).parent / "translations")
TRANSLATION_DIRECTORIES = [  # DEPRECATED: Remove in Kerko 2.x.
    str(pathlib.Path(__file__).parent / "translations")
]

try:
    DEFAULTS = load_toml(pathlib.Path(__file__).parent / "default_config.toml")
except RuntimeError as e:
    print(e, file=sys.stderr)  # noqa: T201
    sys.exit(errno.EINTR)  # This should make the WSGI server exit as well.


def make_blueprint():
    return Blueprint(
        "kerko",
        __name__,
        static_folder="static",
        template_folder="templates",
    )



"""
Functions for extracting data from Zotero items.
"""

import gettext
import itertools
import re
from abc import ABC, abstractmethod
from collections.abc import Iterable
from datetime import datetime

import pycountry
from flask import current_app
from markupsafe import Markup

from kerko.datetime import maximize_partial_date, parse_partial_date
from kerko.richtext import richtext_striptags
from kerko.tags import TagGate
from kerko.text import sort_normalize
from kerko.transformers import find_item_id_in_zotero_uri_links, find_item_id_in_zotero_uris_str

RECORD_SEPARATOR = "\x1e"


def encode_single(value, spec):
    """Encode a single value."""
    return spec.encode(value)


def encode_multiple(value, spec):
    """Encode items of an iterable."""
    return [spec.encode(item) for item in value]


def is_file_attachment(item, mime_types=None):
    """
    Return `True` if a given item is a file attachment.

    :param mime_types: If a list is provided, the item must also match one of
        the given MIME types. If empty or `None`, the MIME type is not checked.
    """
    if not item.get("data"):
        return False
    if not item["data"].get("key"):
        return False
    if item["data"].get("linkMode") not in ["imported_file", "imported_url"]:
        return False
    if mime_types and item["data"].get("contentType", "octet-stream") not in mime_types:
        return False
    return True


def is_link_attachment(item):
    """
    Return `True` if a given item is a link attachment.
    """
    if not item.get("data"):
        return False
    if not item["data"].get("key"):
        return False
    if item["data"].get("linkMode") != "linked_url":
        return False
    if not item["data"].get("url"):
        return False
    return True


class Extractor(ABC):
    """
    Data extractor.

    An extractor can retrieve elements from item or ``LibraryContext`` objects,
    and add elements to a document. The document is represented by a `dict`. A
    ``BaseFieldSpec`` object provides both an ``encode()`` method that may
    transform the data before its assignment into the document, and the key to
    assign the resulting data to.
    """

    def __init__(self, format_="data", encode=encode_single, **kwargs):
        """
        Initialize the extractor.

        :param str format_: Format to retrieve when performing the Zotero item
            read, e.g., 'data', 'bib', 'ris', to ensure that the data required
            by this extractor is requested from Zotero to become available in
            the item at extraction time.

        :param callable encode: Function that can encode a value using a
            ``FieldSpec``.
        """
        self.format = format_
        self.encode = encode
        assert not kwargs  # Subclasses should have consumed every keyword arg.

    @abstractmethod
    def extract(self, item, library_context, spec):
        """
        Retrieve the value from context.

        :return: Extracted value, or `None` if no value could be extracted.
        """

    def extract_and_store(self, document, item, library_context, spec):
        """
        Extract value from context and store its encoded version in document.
        """
        extracted_value = self.extract(item, library_context, spec)
        if extracted_value is not None:
            document[spec.key] = self.encode(extracted_value, spec)

    def warning(self, message, item=None):
        item_ref = f"({item.get('key')})" if item else ""
        current_app.logger.warning(f"{self.__class__.__name__}: {message} {item_ref}")


class TransformerExtractor(Extractor):
    """
    Wrap an extractor to transform data before encoding it into the document.
    """

    def __init__(self, *, extractor, transformers, skip_none_value=True, **kwargs):
        """
        Initialize the extractor.

        :param Extractor extractor: Base extractor to wrap.

        :param list transformers: List of callables that will be chained to
            transform the extracted data. Each callable takes a value as
            argument and returns the transformed value.

        :param bool skip_none_value: If ``true`` (which is the default),
            transformers will not be applied on a ``None`` value.
        """
        super().__init__(format_=extractor.format, **kwargs)
        self.extractor = extractor
        self.transformers = transformers
        self.skip_none_value = skip_none_value

    def apply_transformers(self, value):
        if value is not None or not self.skip_none_value:
            for transformer in self.transformers:
                value = transformer(value)
        return value

    def extract(self, item, library_context, spec):
        value = self.extractor.extract(item, library_context, spec)
        return self.apply_transformers(value)


class MultiExtractor(Extractor):
    """
    Allow a composition of multiple extractors.
    """

    def __init__(self, *, extractors, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)
        self.extractors = extractors

    def extract(self, item, library_context, spec):
        values = []
        for extractor in self.extractors:
            assert self.format == extractor.format  # Extractors can only use same format as parent.
            value = extractor.extract(item, library_context, spec)
            if isinstance(value, Iterable) and not isinstance(value, str):
                values.extend(value)
            elif value:
                values.append(value)
        return values or None


class ChainExtractor(Extractor):
    """
    Extract data using a chain of extractors.

    When the an extractor returns `None`, the following one in the chain is
    tried, until a value is found or no more extractors are left to try in the
    chain.
    """

    def __init__(self, *, extractors, **kwargs):
        super().__init__(**kwargs)
        self.extractors = extractors

    def extract(self, item, library_context, spec):
        value = None
        for extractor in self.extractors:
            assert self.format == extractor.format  # Extractors can only use same format as parent.
            value = extractor.extract(item, library_context, spec)
            if value is not None:
                break
        return value


class KeyExtractor(Extractor):
    def __init__(self, *, key, **kwargs):
        """
        Initialize the extractor.

        :param str key: Key of the element to extract from the Zotero item.
        """
        super().__init__(**kwargs)
        self.key = key


class ItemExtractor(KeyExtractor):
    """Extract a value from an item."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        return item.get(self.key)


class ItemDataExtractor(KeyExtractor):
    """Extract a value from item data."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        return item.get("data", {}).get(self.key)


class ItemTitleExtractor(Extractor):
    """
    Extract the title of an item.

    The name of the field that can be considered the title varies depending on
    the item type ("title", "caseName", "subject", etc.), but in the Zotero
    schema it is always the first field. This extractor uses that premise for
    getting the title, instead of using hardcoded field names.
    """

    def extract(self, item, library_context, spec):  # noqa: ARG002
        item_data = item.get("data", {})
        item_type = item_data.get("itemType")
        if item_type not in ["annotation", "note"] and (
            item_fields := library_context.item_fields.get(item_type)
        ):
            return item_data.get(item_fields[0].get("field"), "")
        return ""


class RawDataExtractor(Extractor):
    def extract(self, item, library_context, spec):  # noqa: ARG002
        return item.get("data")


class ItemRelationsExtractor(Extractor):
    """Extract a list of item's relations corresponding to a given predicate."""

    def __init__(self, predicate, **kwargs):
        super().__init__(**kwargs)
        self.predicate = predicate

    def extract(self, item, library_context, spec):  # noqa: ARG002
        relations = item.get("data", {}).get("relations", {}).get(self.predicate, [])
        if relations and isinstance(relations, str):
            relations = [relations]
        assert isinstance(relations, Iterable)
        return relations


class ItemTypeLabelExtractor(Extractor):
    """Extract the label of the item's type."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        item_type = item.get("data", {}).get("itemType")
        if item_type and item_type in library_context.item_types:
            return library_context.item_types[item_type]
        if item_type != "attachment":  # Attachment has no label, no need to warn.
            self.warning(f"Missing or unknown item type '{item_type}'", item)
        return None


class ItemFieldsExtractor(Extractor):
    """Extract field metadata."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        item_type = item.get("data", {}).get("itemType")
        if item_type and item_type in library_context.item_fields:
            fields = library_context.item_fields[item_type]
            # Retain metadata for fields that are actually present in the item.
            item_fields = [f for f in fields if f.get("field") in item["data"]]
            return item_fields
        if item_type != "attachment":  # Attachment has no field metadata, no need to warn.
            self.warning(f"Missing or unknown item type '{item_type}'", item)
        return None


class ItemLinkExtractor(Extractor):
    """Extract an item link from the 'links' element."""

    def __init__(self, *, link_key, link_type, **kwargs):
        super().__init__(**kwargs)
        self.link_key = link_key
        self.link_type = link_type

    def extract(self, item, library_context, spec):  # noqa: ARG002
        link = item.get("links", {}).get(self.link_key, {})
        if link and link.get("type") == self.link_type:
            return link.get("href")
        return None


class ZoteroWebItemURLExtractor(ItemLinkExtractor):
    """Extract an item's zotero.org link."""

    def __init__(self, *args, **kwargs):
        super().__init__(link_key="alternate", link_type="text/html", *args, **kwargs)


class ZoteroAppItemURLExtractor(Extractor):
    """Extract a link for opening the item in the Zotero app."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        if library_context.library_type == "group":
            return f"zotero://select/groups/{library_context.library_id}/items/{item.get('key')}"
        return f"zotero://select/library/items/{item.get('key')}"


class CreatorTypesExtractor(Extractor):
    """Extract creator types metadata."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        item_type = item.get("data", {}).get("itemType")
        if item_type and item_type in library_context.creator_types:
            library_creator_types = library_context.creator_types[item_type]
            # Retain metadata for creator types that are actually present in the item.
            item_creator_types = []
            for library_creator_type in library_creator_types:
                for item_creator in item["data"].get("creators", []):
                    creator_type = item_creator.get("creatorType")
                    if creator_type and creator_type == library_creator_type.get("creatorType"):
                        item_creator_types.append(library_creator_type)
                        break
            if item_creator_types:
                return item_creator_types
            if item["data"].get("creators", False):
                self.warning(f"Missing creator types for item type '{item_type}'.", item)
        elif item_type != "attachment":  # Attachment has no creator types, no need to warn.
            self.warning(f"Missing or unknown item type '{item_type}'", item)
        return None


class CreatorsExtractor(Extractor):
    """Flatten and extract creator data."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        creators = []
        for creator in item.get("data", {}).get("creators", []):
            fullname = creator.get("name")
            if fullname:
                creators.append(richtext_striptags(fullname).strip())
            firstname = richtext_striptags(creator.get("firstName", "")).strip()
            lastname = richtext_striptags(creator.get("lastName", "")).strip()
            if firstname and lastname:
                # Combine firstname and lastname in different orders to help
                # phrase searches.
                creators.append(f"{firstname} {lastname}")
                creators.append(f"{lastname}, {firstname}")
            elif firstname:
                creators.append(firstname)
            elif lastname:
                creators.append(lastname)
        return RECORD_SEPARATOR.join(creators) if creators else None


class CollectionNamesExtractor(Extractor):
    """Extract item collections for text search."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        names = set()
        for k in item.get("data", {}).get("collections", []):
            if k in library_context.collections:
                name = library_context.collections[k].get("data", {}).get("name", "").strip()
                if name:
                    names.add(name)
        return RECORD_SEPARATOR.join(sorted(names, key=sort_normalize)) if names else None


class BaseTagsExtractor(Extractor):
    def __init__(self, *, include_re="", exclude_re="", **kwargs):
        """
        Initialize the extractor.

        :param str include_re: Any tag that does not matches this regular
            expression will be ignored by the extractor. If empty, all tags will
            be accepted unless `exclude_re` is set and they match it.

        :param str exclude_re: Any tag that matches this regular expression
            will be ignored by the extractor. If empty, all tags will be
            accepted unless `include_re` is set and they do not match it.
        """
        super().__init__(**kwargs)
        self.include = re.compile(include_re) if include_re else None
        self.exclude = re.compile(exclude_re) if exclude_re else None

    def extract(self, item, library_context, spec):  # noqa: ARG002
        tags = set()
        for tag_data in item.get("data", {}).get("tags", []):
            tag = tag_data.get("tag", "").strip()
            if (
                tag
                and (not self.include or self.include.match(tag))
                and (not self.exclude or not self.exclude.match(tag))
            ):
                tags.add(tag)
        return tags or None


class TagsTextExtractor(BaseTagsExtractor):
    """Extract item tags for text search."""

    def extract(self, item, library_context, spec):
        tags = super().extract(item, library_context, spec)
        return RECORD_SEPARATOR.join(sorted(tags, key=sort_normalize)) if tags else None


class LanguageExtractor(Extractor):
    """
    Extract item language(s) into tuples of ISO 639-3 code and language name.

    This uses the language database and translations provided by the pycountry
    library.
    """

    def __init__(
        self,
        *,
        values_separator_re=r";",
        normalize=True,
        locale="en",
        allow_invalid=True,
        normalize_invalid=None,
        **kwargs,
    ):
        """
        Initialize the extractor.

        :param str locale: Locale to translate normalized language names into.
            Ignored when `normalize` is `False`.

        :param str values_separator_re: Regular expression for separating
            multiple values that may have been entered in an item's language
            field.

        :param bool normalize: If `True`, normalize values using the language
            database. If `False`, values are used verbatim.

        :param bool allow_invalid: If `True`, allow values that are not found in
            the language database. Ignored when `normalize` is `False`.

        :param normalize_invalid: Callable to use for normalizing the label when
            the value is invalid. If `None`, then `str.title` will be used.
        """
        super().__init__(encode=encode_multiple, **kwargs)
        self.values_separator = re.compile(values_separator_re)
        self.normalize = normalize
        self.locale = locale
        self.allow_invalid = allow_invalid
        self.normalize_invalid = normalize_invalid or str.title
        self.translations = None
        self.translations_initialized = False

    def extract(self, item, library_context, spec):  # noqa: ARG002
        """
        Extract item language(s) into (value, label) tuples.

        Multiple values are separated using the `self.values_separator` regex.
        """
        values = self.values_separator.split(item.get("data", {}).get("language", ""))
        if self.normalize:
            values = [self.normalize_language(value) for value in values]
        else:
            values = [(value.strip(), value.strip()) for value in values if value]
        # Going through a dict.fromkeys() to eliminate duplicates while preserving ordering.
        return [value for value in dict.fromkeys(values).keys() if value] or None

    def normalize_language(self, value):
        """
        Given a str value, return a corresponding (language code, name) tuple.

        This searches the language database and tries to find an ISO 639-3 code
        corresponding to the given value. If the value has the form "lang-AREA"
        or "lang_AREA", "AREA" is ignored when searching for a language code.
        Matching is case-insensitive and proceeds in the following order,
        stopping at the first match found:

        1. Search for a 3-letter ISO 639-3 code.
        2. Search for a 3-letter ISO 639-2 bibliographic (B) code.
        3. Search for a 2-letter ISO 639-1 code.
        4. Search for an English language name.

        If a matching language is found, a tuple is returned with the 3-letter
        ISO 639-3 code, and the language name. The language name is translated
        in the locale specified by `self.locale`.

        If no matching language is found, a tuple is returned with the value
        converted to lowercase form, and a label. The label is normalized using
        the `self.normalize_case` callable.
        """
        value = value.strip()
        lang = re.split(r"[-_]", value, maxsplit=1)[0]
        match = None
        if len(lang) == 3:  # noqa: PLR2004
            match = pycountry.languages.get(alpha_3=lang)
            if not match:
                match = pycountry.languages.get(bibliographic=lang)
        elif len(lang) == 2:  # noqa: PLR2004
            match = pycountry.languages.get(alpha_2=lang)
        else:
            match = pycountry.languages.get(name=value)
        if match:
            return (match.alpha_3, self.translate_language(match.name))
        if value and self.allow_invalid:
            return (value.lower(), self.normalize_invalid(value))
        return None

    def translate_language(self, name):
        if not self.translations_initialized:
            locale = self.locale.replace("-", "_")
            try:
                if locale.split("_", maxsplit=1)[0].strip() != "en":
                    self.translations = gettext.translation(
                        "iso639-3", pycountry.LOCALES_DIR, languages=[locale]
                    )
            except FileNotFoundError:
                self.warning(f"No language translations found in pycountry for locale '{locale}'.")
            finally:
                self.translations_initialized = True

        if self.translations:
            return self.translations.gettext(name)
        return name


class BaseChildrenExtractor(Extractor):
    def __init__(self, *, item_type, include_re="", exclude_re="", **kwargs):
        """
        Initialize the extractor.

        :param str item_type: The type of child items to extract, either 'note'
            or 'attachment'.

        :param [str,list] include_re: Any child which does not have a tag that
            matches this regular expression will be ignored by the extractor. If
            empty, all children will be accepted unless `exclude_re` is set and
            causes some to be rejected. When passing a list, every pattern of
            the list must match at least a tag for the child to be included.

        :param [str,list] exclude_re: Any child that have a tag that matches
            this regular expression will be ignored by the extractor. If empty,
            all children will be accepted unless `include_re` is set and causes
            some to be rejected. When passing a list, every pattern of the list
            must match at least a tag for the child to be excluded.
        """
        super().__init__(**kwargs)
        self.item_type = item_type
        self.gate = TagGate(include_re, exclude_re)

    def extract(self, item, library_context, spec):  # noqa: ARG002
        accepted_children = []
        for child in item.get("children", []):
            if child.get("data", {}).get("itemType") == self.item_type and self.gate.check(
                child.get("data", {})
            ):
                accepted_children.append(child)
        return accepted_children or None


class BaseChildAttachmentsExtractor(BaseChildrenExtractor):
    def __init__(self, **kwargs):
        super().__init__(item_type="attachment", **kwargs)


class ChildFileAttachmentsExtractor(BaseChildAttachmentsExtractor):
    """
    Extract the metadata of stored copies of files into a list of dicts.
    """

    def __init__(self, *, mime_types=None, **kwargs):
        super().__init__(**kwargs)
        self.mime_types = mime_types

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        return (
            [
                {
                    "id": child["key"],
                    "data": {
                        "contentType": child["data"].get("contentType"),
                        "filename": child["data"].get("filename"),
                        "md5": child["data"].get("md5"),
                        "mtime": child["data"].get("mtime"),
                    },
                }
                for child in children
                if is_file_attachment(child, self.mime_types)
            ]
            if children
            else None
        )


class ChildLinkedURIAttachmentsExtractor(BaseChildAttachmentsExtractor):
    """
    Extract attached links to URIs into a list of dicts.
    """

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        if children:
            return [
                {
                    "title": child["data"].get("title", child["data"].get("url")),
                    "url": child["data"].get("url"),
                }
                for child in children
                if is_link_attachment(child)
            ]
        return None


class ChildAttachmentsFulltextExtractor(BaseChildAttachmentsExtractor):
    """Extract the text content of attachments."""

    def __init__(self, *, mime_types=None, **kwargs):
        super().__init__(**kwargs)
        self.mime_types = mime_types

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        if children:
            return RECORD_SEPARATOR.join(
                [
                    Markup(child["fulltext"]).striptags()
                    for child in children
                    if is_file_attachment(child, self.mime_types) and child.get("fulltext")
                ]
            )
        return None


class BaseChildNotesExtractor(BaseChildrenExtractor):
    def __init__(self, **kwargs):
        super().__init__(item_type="note", **kwargs)


class ChildNotesTextExtractor(BaseChildNotesExtractor):
    """Extract notes for text search."""

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        if children:
            return RECORD_SEPARATOR.join(
                [
                    Markup(child["data"]["note"]).striptags()
                    for child in children
                    if child.get("data", {}).get("note")
                ]
            )
        return None


class RawChildNotesExtractor(BaseChildNotesExtractor):
    """Extract raw notes for storage."""

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        if children:
            return [
                child["data"]["note"] for child in children if child.get("data", {}).get("note")
            ]
        return None


class RelationsInChildNotesExtractor(BaseChildNotesExtractor):
    """Extract item references specified in child notes."""

    def extract(self, item, library_context, spec):
        refs = set()
        children = super().extract(item, library_context, spec)
        if children:
            for child in children:
                note = child.get("data", {}).get("note", "")
                # Find in the href attribute of <a> elements.
                refs.update(find_item_id_in_zotero_uri_links(note))
                # Find in plain text.
                note = Markup(re.sub(r"<br\s*/>", "\n", note)).striptags()
                refs.update(find_item_id_in_zotero_uris_str(note))
        return list(refs) or None


def _expand_paths(path):
    """
    Extract the paths of each of the components of the specified path.

    If the given path is ['a', 'b', 'c'], the returned list of paths is:
    [['a'], ['a', 'b'], ['a', 'b', 'c']]
    """
    return [path[0 : i + 1] for i in range(len(path))]


class CollectionFacetTreeExtractor(Extractor):
    """Index the Zotero item's collections needed for the specified facet."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)

    def extract(self, item, library_context, spec):
        # Sets prevent duplication when multiple collections share common ancestors.
        encoded_ancestors = set()
        for collection_key in item.get("data", {}).get("collections", []):
            if collection_key not in library_context.collections:
                continue  # Skip unknown collection.
            ancestors = library_context.collections.ancestors(collection_key)
            if len(ancestors) <= 1 or ancestors[0] != spec.collection_key:
                continue  # Skip collection, unrelated to this facet.

            ancestors = ancestors[1:]  # Facet values come from subcollections.
            for path in _expand_paths(ancestors):
                label = (
                    library_context.collections.get(path[-1], {})
                    .get("data", {})
                    .get("name", "")
                    .strip()
                )
                encoded_ancestors.add((tuple(path), label))  # Cast path to make it hashable.
        return encoded_ancestors or None


class InCollectionExtractor(Extractor):
    """Extract the boolean membership of an item into a collection."""

    def __init__(self, *, collection_key, true_only=True, check_subcollections=True, **kwargs):
        """
        Initialize the extractor.

        :param str collection_key: Key of the collection to test item membership
            against.

        :param bool true_only: If `True` (default), extraction returns `True`
            when an item belongs to the specified collection, or `None` when it
            does not belong to that collection. If `False`, always return a
            boolean.

        :param bool check_subcollections: If `True` (default), membership is
            extended to any subcollection of the specified collection.
        """
        super().__init__(**kwargs)
        self.collection_key = collection_key
        self.true_only = true_only
        self.check_subcollections = check_subcollections

    def extract(self, item, library_context, spec):  # noqa: ARG002
        item_collections = list(
            itertools.chain(
                *[
                    library_context.collections.ancestors(c) if self.check_subcollections else c
                    for c in item.get("data", {}).get("collections", [])
                ]
            )
        )
        is_in = self.collection_key in item_collections
        if not self.true_only:
            return is_in
        if is_in:
            return True
        return None


class TagsFacetExtractor(BaseTagsExtractor):
    """Index the Zotero item's tags for faceting."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)


class ItemTypeFacetExtractor(Extractor):
    """Index the Zotero item's type for faceting."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        item_type = item.get("data", {}).get("itemType")
        if item_type:
            return (item_type, library_context.item_types.get(item_type, item_type))
        self.warning("Missing itemType", item)
        return None


class YearExtractor(Extractor):
    """Parse the Zotero item's publication date to get just the year."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        parsed_date = item.get("meta", {}).get("parsedDate", "")
        if parsed_date:
            year, _month, _day = parse_partial_date(parsed_date)
            return str(year)
        return None


class YearFacetExtractor(Extractor):
    """Index the Zotero item's publication date for faceting by year."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)

    def extract(self, item, library_context, spec):  # noqa: ARG002
        parsed_date = item.get("meta", {}).get("parsedDate", "")
        if parsed_date:
            year, _month, _day = parse_partial_date(parsed_date)
            decade = int(int(year) / 10) * 10
            century = int(int(year) / 100) * 100
            return _expand_paths([str(century), str(decade), str(year)])
        return None


class ItemDataLinkFacetExtractor(ItemDataExtractor):
    def extract(self, item, library_context, spec):  # noqa: ARG002
        return item.get("data", {}).get(self.key, "").strip() != ""


class MaximizeParsedDateExtractor(Extractor):
    """Extract and "maximize" a `datetime` object from the item's `parsedDate` meta field."""

    def extract(self, item, library_context, spec):  # noqa: ARG002
        parsed_date = item.get("meta", {}).get("parsedDate", None)
        if parsed_date:
            try:
                return datetime(*maximize_partial_date(*parse_partial_date(parsed_date)))
            except ValueError:
                pass
        return None


def _prepare_sort_text(text):
    """
    Normalize the given text for a sort field.

    :param str text: The Unicode string to normalize.

    :return bytearray: The normalized text.
    """
    return sort_normalize(Markup(text).striptags())


class SortItemDataExtractor(ItemDataExtractor):
    def extract(self, item, library_context, spec):
        return _prepare_sort_text(super().extract(item, library_context, spec))


class SortTitleExtractor(ItemTitleExtractor):
    def extract(self, item, library_context, spec):
        return _prepare_sort_text(super().extract(item, library_context, spec))


class SortCreatorExtractor(Extractor):
    def extract(self, item, library_context, spec):  # noqa: ARG002
        creators = []

        def append_creator(creator):
            creator_parts = [
                _prepare_sort_text(creator.get("lastName", "")),
                _prepare_sort_text(creator.get("firstName", "")),
                _prepare_sort_text(creator.get("name", "")),
            ]
            creators.append(" zzz ".join([p for p in creator_parts if p]))

        # We treat creator types like an ordered list, where the first creator
        # type is for primary creators. Depending on the citation style, lesser
        # creator types may not appear in citations. Therefore, we try to sort
        # only by primary creators in order to avoid sorting with data that may
        # be invisible to the user. Only when an item has no primary creator do
        # we fallback to lesser creators.
        for creator_type in library_context.get_creator_types(item.get("data", {})):
            for creator in item.get("data", {}).get("creators", []):
                if creator.get("creatorType", "") == creator_type.get("creatorType"):
                    append_creator(creator)
            if creators:
                break  # No need to include lesser creator types.
        return " zzzzzz ".join(creators)


class SortDateExtractor(Extractor):
    def extract(self, item, library_context, spec):  # noqa: ARG002
        parsed_date = item.get("meta", {}).get("parsedDate", "")
        year, month, day = parse_partial_date(parsed_date)
        return int(f"{year:04d}{month:02d}{day:02d}")



from flask import url_for

from kerko.richtext import richtext_striptags
from kerko.shortcuts import composer


def format_creator_name(creator):
    """Format a name from a dict having either a 'name' key, or 'firstName' and 'lastName' keys."""
    return creator.get(
        "name",
        ", ".join(
            [
                name_part
                for name_part in [
                    creator.get("lastName", "").strip(),
                    creator.get("firstName", "").strip(),
                ]
                if name_part
            ]
        ),
    )


def inject_creator_display_names(item, link=True):
    """
    Inject the display names of creators into the given item.

    :param dict item: The item whose creators are to be updated.

    :param bool link: Whether to provide a search link for the creator.
    """
    for spec in item.keys():
        if spec == "data" and "creators" in item["data"]:
            for creator in item["data"]["creators"]:
                # Add creator display name.
                creator["display"] = format_creator_name(creator)
                # Add creator type labels.
                if "creator_types" in item:
                    for t in item["creator_types"]:
                        if t["creatorType"] == creator["creatorType"]:
                            creator["label"] = t["localized"]
                            break
                if link:
                    creator_scope = composer().scopes.get("creator")
                    if creator_scope:
                        creator["url"] = url_for(
                            ".search",
                            **creator_scope.add_keywords(
                                value=f'"{richtext_striptags(creator["display"])}"'
                            ),
                        )



"""
Unit tests for the tags module.
"""

import unittest

from kerko.tags import TagGate


class TagGateTestCase(unittest.TestCase):
    def setUp(self):
        self.objects = [
            {  # 0
                "tags": [
                    {"tag": "_include"},
                ]
            },
            {  # 1
                "tags": [
                    {"tag": "foo"},
                    {"tag": "bar"},
                    {"tag": "_include"},
                    {"tag": "baz"},
                    {"tag": "qux"},
                ]
            },
            {  # 2
                "tags": [
                    {"tag": "_exclude"},
                ]
            },
            {  # 3
                "tags": [
                    {"tag": "foo"},
                    {"tag": "bar"},
                    {"tag": "_exclude"},
                    {"tag": "baz"},
                    {"tag": "qux"},
                ]
            },
            {  # 4
                "tags": [
                    {"tag": "_include"},
                    {"tag": "_exclude"},
                ]
            },
            {  # 5
                "tags": []
            },
            {  # 6
                "tags": [
                    {"tag": "foo"},
                    {"tag": "bar"},
                ]
            },
            {  # 7
                "tags": [
                    {"tag": "foo"},
                    {"tag": "bar"},
                    {"tag": "_include1"},
                    {"tag": "baz"},
                    {"tag": "_include2"},
                    {"tag": "_include3"},
                    {"tag": "qux"},
                ]
            },
            {  # 8
                "tags": [
                    {"tag": "foo"},
                    {"tag": "bar"},
                    {"tag": "baz"},
                    {"tag": "qux"},
                    {"tag": "_exclude1"},
                    {"tag": "_exclude2"},
                ]
            },
            {  # 9
                "tags": [
                    {"tag": "_exclude1"},
                    {"tag": "_include1"},
                    {"tag": "_exclude2"},
                    {"tag": "foo"},
                    {"tag": "bar"},
                    {"tag": "_include2"},
                    {"tag": "baz"},
                    {"tag": "qux"},
                    {"tag": "_include3"},
                ]
            },
        ]

    def test_default(self):
        gate = TagGate()
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertTrue(gate.check(self.objects[2]))
        self.assertTrue(gate.check(self.objects[3]))
        self.assertTrue(gate.check(self.objects[4]))
        self.assertTrue(gate.check(self.objects[5]))
        self.assertTrue(gate.check(self.objects[6]))
        self.assertTrue(gate.check(self.objects[7]))
        self.assertTrue(gate.check(self.objects[8]))
        self.assertTrue(gate.check(self.objects[9]))

    def test_match_include(self):
        gate = TagGate(include_re=r"^_include$")
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertTrue(gate.check(self.objects[4]))
        self.assertFalse(gate.check(self.objects[5]))
        self.assertFalse(gate.check(self.objects[6]))
        self.assertFalse(gate.check(self.objects[7]))
        self.assertFalse(gate.check(self.objects[8]))
        self.assertFalse(gate.check(self.objects[9]))

    def test_match_include_wildcard(self):
        gate = TagGate(include_re=r"^_incl.*$")
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertTrue(gate.check(self.objects[4]))
        self.assertFalse(gate.check(self.objects[5]))
        self.assertFalse(gate.check(self.objects[6]))
        self.assertTrue(gate.check(self.objects[7]))
        self.assertFalse(gate.check(self.objects[8]))
        self.assertTrue(gate.check(self.objects[9]))

    def test_match_include_multiple_tags(self):
        gate = TagGate(include_re=[r"^_include1$", r"^_include2$", r"^_include3$"])
        self.assertFalse(gate.check(self.objects[0]))
        self.assertFalse(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertFalse(gate.check(self.objects[4]))
        self.assertFalse(gate.check(self.objects[5]))
        self.assertFalse(gate.check(self.objects[6]))
        self.assertTrue(gate.check(self.objects[7]))
        self.assertFalse(gate.check(self.objects[8]))
        self.assertTrue(gate.check(self.objects[9]))

    def test_match_exclude(self):
        gate = TagGate(exclude_re=r"^_exclude$")
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertFalse(gate.check(self.objects[4]))
        self.assertTrue(gate.check(self.objects[5]))
        self.assertTrue(gate.check(self.objects[6]))
        self.assertTrue(gate.check(self.objects[7]))
        self.assertTrue(gate.check(self.objects[8]))
        self.assertTrue(gate.check(self.objects[9]))

    def test_match_exclude_wildcard(self):
        gate = TagGate(exclude_re=r"^_excl.*$")
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertFalse(gate.check(self.objects[4]))
        self.assertTrue(gate.check(self.objects[5]))
        self.assertTrue(gate.check(self.objects[6]))
        self.assertTrue(gate.check(self.objects[7]))
        self.assertFalse(gate.check(self.objects[8]))
        self.assertFalse(gate.check(self.objects[9]))

    def test_match_exclude_multiple_tags(self):
        gate = TagGate(exclude_re=[r"^_exclude1$", r"^_exclude2$"])
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertTrue(gate.check(self.objects[2]))
        self.assertTrue(gate.check(self.objects[3]))
        self.assertTrue(gate.check(self.objects[4]))
        self.assertTrue(gate.check(self.objects[5]))
        self.assertTrue(gate.check(self.objects[6]))
        self.assertTrue(gate.check(self.objects[7]))
        self.assertFalse(gate.check(self.objects[8]))
        self.assertFalse(gate.check(self.objects[9]))

    def test_match_both(self):
        gate = TagGate(include_re=r"^_include$", exclude_re=r"^_exclude$")
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertFalse(gate.check(self.objects[4]))
        self.assertFalse(gate.check(self.objects[5]))
        self.assertFalse(gate.check(self.objects[6]))
        self.assertFalse(gate.check(self.objects[7]))
        self.assertFalse(gate.check(self.objects[8]))
        self.assertFalse(gate.check(self.objects[9]))

    def test_match_both_multiple_tags(self):
        gate = TagGate(
            include_re=[r"^_include1$", r"^_include2$", r"^_include3$"],
            exclude_re=[r"^_exclude1$", r"^_exclude2$"],
        )
        self.assertFalse(gate.check(self.objects[0]))
        self.assertFalse(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertFalse(gate.check(self.objects[4]))
        self.assertFalse(gate.check(self.objects[5]))
        self.assertFalse(gate.check(self.objects[6]))
        self.assertTrue(gate.check(self.objects[7]))
        self.assertFalse(gate.check(self.objects[8]))
        self.assertFalse(gate.check(self.objects[9]))



"""
Dedicated container to display current search criteria.

The name "breadbox" is a reference to another well-known navigational aid
user interface pattern: the "breadcrumbs".
"""

from collections import defaultdict
from copy import copy

from flask import url_for

from kerko.shortcuts import composer


def _build_keywords(criteria):
    """Build the active keywords in the same format as a facet."""
    keywords = defaultdict(list)
    for scope in composer().scopes.values():
        if scope.key in criteria.keywords:
            for value in criteria.keywords.getlist(scope.key):
                keywords[scope.key].append(
                    {
                        "label": value,
                        "remove_url": url_for(
                            ".search",
                            **criteria.params(
                                keywords=scope.remove_keywords(value, criteria.keywords),
                                options={
                                    "page": None,
                                    "page-len": None,
                                    "id": None,
                                },
                            ),
                        ),
                    }
                )
    return keywords


def _get_active_filters(results):
    filters = []
    for item in results:
        if item.get("remove_url", None):
            if item.get("children", []):
                item = copy(item)  # noqa: PLW2901
                item["children"] = _get_active_filters(item["children"])
            filters.append(item)
    return filters


def _build_filters(facet_results):
    """Collect just the active facets from the given facet results."""
    filters = {}
    for key, results in facet_results.items():
        facet_filters = _get_active_filters(results)
        if facet_filters:
            filters[key] = facet_filters
    return filters


def build_breadbox(criteria, facet_results):
    breadbox = {}

    breadbox_keywords = _build_keywords(criteria)
    if breadbox_keywords:
        breadbox["keywords"] = breadbox_keywords

    breadbox_filters = _build_filters(facet_results)
    if breadbox_filters:
        breadbox["filters"] = breadbox_filters

    return breadbox



"""
Utilities related to exceptions.
"""

import wrapt
from flask import abort, current_app


def except_abort(exc, status):
    """
    Decorate a function in order to call `abort()` upon catching an exception.

    This also logs an error.

    :param Exception exc: Exception (or tuple of exceptions) to catch.

    :param int status: HTTP response status code to return in case of an exception.
    """

    @wrapt.decorator
    def wrapper(wrapped, _instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except exc as e:
            current_app.logger.error(e)
            abort(status)

    return wrapper


def except_raise(catch_exc, raise_exc, raise_args):
    """
    Decorate a function to raise an exception upon catching a given exception.

    This also logs an error.

    :param Exception catch_exc: Exception (or tuple of exceptions) to catch.

    :param Exception raise_exc: Exception to raise.

    :param raise_args: Arguments to pass to the exception constructor.
    """

    @wrapt.decorator
    def wrapper(wrapped, _instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except catch_exc as e:
            current_app.logger.error(e)
            raise raise_exc(raise_args) from e

    return wrapper



"""Item view building functions."""

from urllib.parse import urlparse

from flask import request, url_for

from kerko.richtext import richtext_escape
from kerko.shortcuts import config
from kerko.views.item import creators, facets, meta, relations


def is_valid_url(url):
    parsed_url = urlparse(url)
    return parsed_url.scheme and parsed_url.netloc


def inject_item_data(item):
    creators.inject_creator_display_names(item)
    relations.inject_relations(item)
    facets.inject_facet_results(item)


def get_item_title(item):
    """
    Extract the title of an item.

    The name of the field that can be considered the title varies depending on
    the item type ("title", "caseName", "subject", etc.), but in the Zotero
    schema it is always the first field. This extractor uses that premise for
    getting the title (instead of using hardcoded field names).
    """
    item_data = item.get("data", {})
    if item_data.get("itemType") not in ["annotation", "note"] and (
        item_fields := item.get("item_fields")
    ):
        return item_data.get(item_fields[0].get("field"), "")
    return ""


def build_item_context(item):
    context = {
        "item": item,
        "item_url": url_for(".item_view", item_id=item["id"], _external=True),
        "title": richtext_escape(get_item_title(item)),
        "highwirepress_tags": meta.build_highwirepress_tags(item),
    }
    if config("kerko.features.open_in_zotero_app"):
        context["open_in_zotero_app"] = bool(request.cookies.get("open-in-zotero-app"))
        url = item.get("zotero_app_url", "")
        if url and is_valid_url(url):
            context["open_in_zotero_app_url"] = url
    if config("kerko.features.open_in_zotero_web"):
        context["open_in_zotero_web"] = bool(request.cookies.get("open-in-zotero-web"))
        url = item.get("zotero_web_url", "")
        if url and is_valid_url(url):
            context["open_in_zotero_web_url"] = url
    return context



from kerko.searcher import Searcher
from kerko.shortcuts import composer, config
from kerko.storage import open_index


def _get_forward_rel_filters(item, relation_spec):
    """
    Build search terms for a forward relation.

    Finds items for which one of the identifiers match an identifier from the
    current item's relation field.
    """
    return {
        field.key: [ref for ref in item.get(relation_spec.field.key, [])]
        for field in relation_spec.id_fields
    }


def _get_reverse_rel_filters(item, relation_spec):
    """
    Build search terms for a reverse relation.

    Finds occurrences of one of the identifiers of the current item in the
    relation field of other items.
    """
    identifiers = []
    for id_field in relation_spec.id_fields:
        value = item.get(id_field.key, [])
        if isinstance(value, str):
            identifiers.append(value)
        else:
            identifiers.extend(value)
    return {relation_spec.field.key: [identifier for identifier in identifiers]}


def _get_bidirectional_rel_filters(item, relation_spec):
    return {
        **_get_forward_rel_filters(item, relation_spec),
        **_get_reverse_rel_filters(item, relation_spec),
    }


def inject_relations(item):
    """
    Search and inject relations into the given item.
    """
    common_search_args = {
        "limit": None,
        "reject_any": {"item_type": ["note", "attachment"]},
        "sort_spec": composer().sorts.get(config("kerko.features.relations_sort")),
        "faceting": False,
    }
    # For each related item, load the same fields as in normal search result
    # lists, except for the COinS field because we don't want it to get rendered
    # when displaying relations.
    related_item_fields = composer().select_fields(
        [
            key
            for key in config("kerko.search.result_fields")
            + [badge.field.key for badge in composer().badges.values()]
            if key != "coins"
        ],
    )
    index = open_index("index")
    with Searcher(index) as searcher:
        for relation_spec in composer().relations.values():
            if relation_spec.directed:
                item[relation_spec.field.key] = searcher.search(
                    require_any=_get_forward_rel_filters(item, relation_spec),
                    **common_search_args,
                ).items(related_item_fields)
                if relation_spec.reverse:
                    item[relation_spec.reverse_field_key] = searcher.search(
                        require_any=_get_reverse_rel_filters(item, relation_spec),
                        **common_search_args,
                    ).items(related_item_fields)
            else:
                item[relation_spec.field.key] = searcher.search(
                    require_any=_get_bidirectional_rel_filters(item, relation_spec),
                    **common_search_args,
                ).items(related_item_fields)



"""
Date and time utilities.
"""

import re
from calendar import monthrange
from datetime import datetime

from babel import dates
from flask_babel import get_locale, get_timezone


def parse_partial_date(text, default_year=0, default_month=0, default_day=0):
    """
    Parse a partial date into a (year, month, day) tuple of numerical values.

    If year, month, and/or day is missing, replace it with the specified default
    value.
    """
    year, month, day = default_year, default_month, default_day
    matches = re.match(r"^([0-9]{4})(-([0-9]{2})(-([0-9]{2}))?)?", text)
    if matches:
        if matches.group(1):
            year = int(matches.group(1))
        if matches.group(3):
            month = int(matches.group(3))
        if matches.group(5):
            day = int(matches.group(5))
    return year, month, day


def maximize_partial_date(year, month, day):
    """
    Maximize the values of a (potentially) partial date.

    If `year`, `month`, and/or `day` is `0`, replace with the maximum calendar
    value. If all are `0`, return today's date. If the known date parts match
    today's corresponding date parts, complete the date with today's parts
    instead of the maximum calendar value.
    """
    today = datetime.today()
    if year == 0:
        year, month, day = today.year, today.month, today.day
    if month == 0:
        if year == today.year:
            month, day = today.month, today.day
        else:
            month = 12
    if day == 0:
        if year == today.year and month == today.month:
            day = today.day
        else:
            day = monthrange(year, month)[1]  # Use last day of the month.
    return year, month, day


def reformat_date(value, **kwargs):
    """
    Reformat a date-time string into a localized form.

    A strict parsing of the value is performed, based on the ISO 8601 format
    (calendar dates with time and timezone designator are supported, other
    variants are not). If the value cannot be parsed, the original value is
    returned as is. Date values in Zotero are often manually entered by the
    user, lacking parts such as the day, the time, or the timezone; those values
    will be returned as is.

    :param str value: The value to parse.

    :param dict kwargs: Parameters to pass along to `format_datetime`, once the
        value is parsed.

    :return str: The localized version of the date value, if `value` could be
        parsed, otherwise the original `value` is returned as is.
    """

    # Note: Full ISO 8601 support with strptime was not possible before Python
    # 3.7, because the %z directive didn't support the colon separator.
    # Reference: https://docs.python.org/3/library/datetime.html#technical-detail
    formats_to_try = [
        "%Y-%m-%dT%H:%M:%S.%fZ",
        "%Y-%m-%dT%H:%M:%S.%f%z",
        "%Y-%m-%dT%H:%M:%SZ",
        "%Y-%m-%dT%H:%M:%S%z",
        "%Y-%m-%dT%H:%MZ",
        "%Y-%m-%dT%H:%M%z",
        "%Y-%m-%dT%HZ",
        "%Y-%m-%dT%H%z",
    ]

    for fmt in formats_to_try:
        try:
            parsed = datetime.strptime(value, fmt)
        except ValueError:
            pass
        else:
            return format_datetime(parsed, **kwargs)
    return value


def format_datetime(dt, *, convert_tz=False, show_tz=False):
    """
    Format a `datetime` object into a localized form.

    :param datetime dt: The datetime object to format.

    :param bool convert_tz: Whether to convert the value to Babel's timezone.

    :param bool show_tz: Whether to print the timezone's name in the result.

    :return str: The localized version of the datetime object.
    """
    parts = [
        dates.format_datetime(
            dt, format="short", tzinfo=get_timezone() if convert_tz else None, locale=get_locale()
        ),
    ]
    if show_tz:
        if convert_tz:
            parts.append(f"({dt.astimezone(get_timezone()).tzname()})")
        else:
            parts.append(f"({dt.tzname()})")
    return " ".join(parts)


def iso_to_timestamp(date_str):
    """
    Convert an ISO 8601 date string to a timestamp.

    Note: This does not accept all ISO 8601 strings, but just the following
    format: %Y-%m-%dT%H:%M:%SZ

    :return int: Timestamp rounded to an integer.
    """
    return round(iso_to_datetime(date_str).timestamp())


def iso_to_datetime(date_str):
    """
    Convert an ISO 8601 date string to a `datetime` object.

    Note: This does not accept all ISO 8601 strings, but just the following
    format: %Y-%m-%dT%H:%M:%SZ

    :return datetime: `datetime` object.
    """
    return datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ")
    # Note: Using strptime() instead of fromisoformat(), because the latter only
    # accepts the specific string format since Python 3.11+.



from time import sleep

import requests
import wrapt
from flask import current_app
from pyzotero import zotero, zotero_errors

from kerko.shortcuts import config


class LibraryContext:
    """Contains data related to a Zotero library."""

    def __init__(
        self,
        library_id,
        library_type,
        *,
        collections,
        item_types,
        item_fields,
        creator_types,
    ):
        self.library_id = library_id
        self.library_type = library_type
        self.collections = collections
        self.item_types = item_types
        self.item_fields = item_fields
        self.creator_types = creator_types

    def get_creator_types(self, item_data):
        return self.creator_types.get(item_data.get("itemType", ""), [])


@wrapt.decorator
def retry_zotero(wrapped, _instance, args, kwargs):
    """
    Retry the wrapped function if the Zotero API call fails.

    Caution: This decorator should only be used on simple functions, as the
    whole function is called repeatedly as long a Zotero fails.
    """
    attempts = 1
    while True:
        try:
            return wrapped(*args, **kwargs)
        except (
            requests.exceptions.ConnectionError,
            zotero_errors.HTTPError,
            zotero_errors.UnsupportedParams,
        ) as e:
            current_app.logger.warning(e)
            if attempts < config("kerko.zotero.max_attempts"):
                current_app.logger.warning(
                    f"The Zotero API request has failed in {wrapped.__name__}. "
                    f"New attempt in {config('kerko.zotero.wait')} seconds..."
                )
                attempts += 1
                sleep(config("kerko.zotero.wait"))
            else:
                current_app.logger.error(
                    "The maximum number of API call attempts to Zotero has been reached. Stopping."
                )
                raise


def init_zotero():
    return zotero.Zotero(
        library_id=config("ZOTERO_LIBRARY_ID"),
        library_type=config("ZOTERO_LIBRARY_TYPE"),
        api_key=config("ZOTERO_API_KEY"),
        locale=config("kerko.zotero.locale"),
    )


def request_library_context(zotero_credentials):
    item_types = {t["itemType"]: t["localized"] for t in load_item_types(zotero_credentials)}
    return LibraryContext(
        zotero_credentials.library_id,
        zotero_credentials.library_type.rstrip("s"),  # Remove 's' appended by pyzotero.
        collections=Collections(zotero_credentials),
        item_types=item_types,
        item_fields={t: load_item_type_fields(zotero_credentials, t) for t in item_types.keys()},
        creator_types={
            t: load_item_type_creator_types(zotero_credentials, t) for t in item_types.keys()
        },
    )


@retry_zotero
def last_modified_version(zotero_credentials):
    return zotero_credentials.last_modified_version()


@retry_zotero
def load_item(zotero_credentials, item_key):
    """Return a specific item."""
    return zotero_credentials.item(item_key)


@retry_zotero
def load_item_types(zotero_credentials):
    """Return all Zotero item types."""
    current_app.logger.info("Requesting all item types...")
    return zotero_credentials.item_types()


@retry_zotero
def load_item_fields(zotero_credentials):
    """Return all Zotero fields."""
    current_app.logger.info("Requesting all fields...")
    return zotero_credentials.item_fields()


@retry_zotero
def load_item_type_fields(zotero_credentials, item_type):
    """Return all Zotero fields for a given item type."""
    current_app.logger.info(f"Requesting fields for items of type '{item_type}'...")
    return zotero_credentials.item_type_fields(item_type)


@retry_zotero
def load_item_type_creator_types(zotero_credentials, item_type):
    """List all Zotero creator types for a given item type."""
    current_app.logger.info(f"Requesting creator types for items of type '{item_type}'...")
    return zotero_credentials.item_creator_types(item_type)


@retry_zotero
def load_deleted_or_trashed_items(zotero_credentials, since):
    deleted = zotero_credentials.deleted(since=since).get("items", [])
    trashed = [trashed["key"] for trashed in Items(zotero_credentials, since=since, trash=True)]
    return deleted + trashed


@retry_zotero
def load_new_fulltext(zotero_credentials, since):
    current_app.logger.info(f"Requesting updated text content since version {since}...")
    items = zotero_credentials.new_fulltext(since)
    current_app.logger.info(f"Found {len(items)} item(s) with updated text content.")
    return items


@retry_zotero
def load_item_fulltext(zotero_credentials, item_key):
    current_app.logger.debug(f"Requesting text content of item {item_key}...")
    try:
        response = zotero_credentials.fulltext_item(item_key)
        if response.get("content") and (
            response.get("indexedChars", 0) > 0 or response.get("indexedPages", 0) > 0
        ):
            return response["content"]
        current_app.logger.info(f"Text content empty for item {item_key}.")
    except zotero_errors.ResourceNotFound:
        current_app.logger.info(f"Text content not available for item {item_key}.")
    return None


@retry_zotero
def retrieve_file(zotero_credentials, item_id):
    """Retrieve the file from the given item."""
    return zotero_credentials.file(item_id)


class Collections:
    """
    Iterable over all collections found in Zotero.

    Collections are loaded in memory all at once, and may be acccessed by key
    using the dict[] notation.

    Each collection is a dict as returned by Zotero. Example::

        {
            'data': {
                'key': 'C435CGN4',
                'name': 'Example Collection Name',
                'parentCollection': 'D5U2TEVB',
                'relations': {   },
                'version': 114
            },
            'key': 'C435CGN4',
            'library': {
                'id': 2076305,
                'links': {
                    'alternate': {
                        'href': 'https://www.example.com',
                        'type': 'text/html'
                    }
                },
                'name': 'Example Library',
                'type': 'user'
            },
            'links': {
                'alternate': {
                    'href': 'https://www.zotero.org/example_library/collections/C435CGN4',
                    'type': 'text/html'
                },
                'self': {
                    'href': 'https://api.zotero.org/users/9999999/collections/C435CGN4',
                    'type': 'application/json'
                },
                'up': {
                    'href': 'https://api.zotero.org/users/9999999/collections/D5U2TEVB',
                    'type': 'application/atom+xml'
                }
            },
            'meta': {
                'numCollections': 0,
                'numItems': 15
            },
            'version': 114
        }
    """

    def __init__(self, zotero_credentials, top_level=False):
        self.collections = {}
        # Immediately load all collections, to allow later access by key.
        self.load_collections(zotero_credentials, top_level)
        self.iterator = None

    @retry_zotero
    def load_collections(self, zotero_credentials, top_level):
        current_app.logger.info(
            "Requesting {which} collections...".format(which="top-level" if top_level else "all")
        )
        start = 0
        if top_level:
            method = zotero_credentials.collections_top
        else:
            method = zotero_credentials.collections
        while True:
            more = method(start=start, limit=config("kerko.zotero.batch_size"))
            if not more:
                break
            start += len(more)
            for collection in more:
                self.collections[collection["key"]] = collection

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.collections)

    def __getitem__(self, key):
        return self.collections[key]

    def __contains__(self, key):
        return key in self.collections

    def __next__(self):
        if self.iterator is None:
            self.iterator = iter(self.collections.values())
        return next(self.iterator)

    def get(self, key, default):
        return self.collections.get(key, default)

    def ancestors(self, key):
        """
        Return the ancestors of the specified collection.

        :return: A list representing the path of the specified collection,
            including the specified collection. Each list element is a
            collection key.
        """
        collection = self.collections.get(key)
        if collection:
            parent_key = collection["data"].get("parentCollection")
            if parent_key:
                ancestors = self.ancestors(parent_key)
                ancestors.append(key)
                return ancestors
            return [key]
        return []


class Items:
    """
    Iterable over Zotero items.

    Items are loaded on demand, in small batches, and cannot be accessed
    directly, but can be iterated on. Each item is represented by a (item,
    children) tuple, where item is a dict as returned by Zotero, and chidren a
    list of the child items (also dicts as returned by Zotero).
    """

    def __init__(self, zotero_credentials, *, since=0, formats=None, trash=False):
        """
        Construct the iterable.

        :param zotero.Zotero zotero_credentials: The Zotero instance.

        :param int since: Retrieve only items modified after this library
            version.

        :param Iterable formats: Iterable of format values for the Zotero read.
            Defaults to `['data']`. See available formats on
            https://www.zotero.org/support/dev/web_api/v3/basics#parameters_for_format_json.

        :param bool Trash: If `False`, all library items except those in the
            trash are returned. If `True`, only items from the trash are
            returned.
        """
        self.zotero_credentials = zotero_credentials
        self.since = since
        self.include = ",".join(sorted(formats or ["data"]))
        if trash:
            self.method = "trash"
            self.method_info = "trashed"
        else:
            self.method = "items"
            self.method_info = "updated"
        self.start = 0
        self.zotero_batch = []
        self.iterator = iter(self.zotero_batch)

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            try:
                return self._next_item()
            except StopIteration:
                self._next_batch()

    @retry_zotero
    def _next_batch(self):
        limit = config("kerko.zotero.batch_size")
        current_app.logger.info(
            f"Requesting up to {limit} {self.method_info} items since version {self.since}, "
            f"starting at position {self.start}..."
        )
        params = {
            "since": self.since,
            "start": self.start,
            "limit": limit,
            "sort": "dateAdded",
            "direction": "asc",
            "include": self.include,
            "style": config("kerko.zotero.csl_style"),
        }
        self.zotero_batch = getattr(self.zotero_credentials, self.method)(**params)
        if not self.zotero_batch:
            raise StopIteration  # Empty batch, nothing more to iterate on.
        self.iterator = iter(self.zotero_batch)

    def _next_item(self):
        zotero_item = next(self.iterator)
        self.start += 1
        return zotero_item



from kerko.criteria import create_search_criteria
from kerko.shortcuts import composer


def inject_facet_results(item):
    """
    Prepare facet "results" corresponding to the given item's faceting fields.

    The only distinction from real facet results obtained from a search is that
    counts are zero. Using the same data structure as real facet results allows
    the reuse of facet display logic.
    """
    item["facet_results"] = {}
    for spec_key in composer().facets.keys() & item.keys():
        if isinstance(item[spec_key], list):
            fake_results = {value: 0 for value in item[spec_key]}
        else:
            fake_results = {item[spec_key]: 0}
        # Use empty criteria -- the facets will provide starting points for new searches.
        item["facet_results"][spec_key] = (
            composer().facets[spec_key].build(fake_results, criteria=create_search_criteria())
        )



"""Utilities for preparing embedded metadata for HTML pages."""

import re

from flask import url_for

from kerko.shortcuts import config
from kerko.views.item.creators import format_creator_name


def build_highwirepress_tags(item):
    # References:
    # - https://scholar.google.ca/intl/en/scholar/inclusion.html#indexing
    # - https://www.monperrus.net/martin/accurate+bibliographic+metadata+and+google+scholar
    tags = []
    data = item.get("data", {})
    if config("kerko.meta.highwirepress_tags") and data.get("itemType") in [
        "book",
        "conferencePaper",
        "journalArticle",
        "report",
        "thesis",
    ]:
        tags.append(("citation_title", data.get("title", data.get("shortTitle", "")).strip()))
        tags.append(("citation_publication_date", data.get("date", "")))
        tags.append(("citation_date", data.get("date", "")))
        tags.append(("citation_year", item.get("year", "")))

        for creator in data.get("creators", []):
            if creator.get("creatorType", "") == "author":
                # Google Scholar requires that only the actual authors be included.
                tags.append(("citation_author", format_creator_name(creator)))

        for tag in ["volume", "issue", "language", "publisher"]:
            if data.get(tag):
                tags.append((f"citation_{tag.lower()}", data[tag]))

        pages = data.get("pages")
        if pages:
            pages = re.sub("[-\u2013]", "-", pages)  # Replace hyphens and en-dashes.
            pages = [p.strip() for p in re.split(r"\W*-\W*", data["pages"], maxsplit=2)]
            if pages and pages[0]:
                tags.append(("citation_firstpage", pages[0]))
            if len(pages) > 1 and pages[1]:
                tags.append(("citation_lastpage", pages[1]))

        if data["itemType"] == "conferencePaper":
            conference_title = data.get("conferenceName", "").strip()
            if conference_title:
                tags.append(("citation_conference_title", conference_title))
        elif data["itemType"] == "journalArticle":
            publication_title = data.get(
                "publicationTitle",
                data.get("journalAbbreviation", ""),
            ).strip()
            if publication_title:
                tags.append(("citation_journal_title", publication_title))
        elif data["itemType"] == "thesis":
            university = data.get("university", "").strip()
            if university:
                tags.append(("citation_dissertation_institution", university))
        elif data["itemType"] == "report":
            institution = data.get("institution", "").strip()
            if institution:
                tags.append(("citation_technical_report_institution", institution))
            report_number = data.get("reportNumber", "").strip()
            if report_number:
                tags.append(("citation_technical_report_number", report_number))

        for tag in ["ISBN", "ISSN", "DOI"]:
            if data.get(tag):
                tags.append((f"citation_{tag.lower()}", data[tag]))
            elif data.get("extra"):  # Look for it in the Extra field.
                matches = re.search(
                    rf"^\s*({tag}):\s*(\S+)\s*$", data["extra"], re.IGNORECASE | re.MULTILINE
                )
                if matches:
                    tags.append((f"citation_{tag.lower()}", matches[2]))

        for attachment in item.get("attachments", []):
            if attachment.get("data", {}).get("contentType") == "application/pdf":
                tags.append(
                    (
                        "citation_pdf_url",
                        url_for(
                            ".child_attachment_download",
                            item_id=item["id"],
                            attachment_id=attachment["id"],
                            attachment_filename=attachment["data"].get("filename", item["id"]),
                            _external=True,
                        ),
                    )
                )
                break  # Only attach the first PDF found.

    return tags



from werkzeug.datastructures import MultiDict

from kerko.shortcuts import composer, config


def create_search_criteria(initial=None):
    return Criteria(
        initial=initial,
        options_initializers=[
            Criteria.initialize_page,
            Criteria.initialize_page_len,
            Criteria.initialize_sort,  # Must run after initialize_keywords().
            Criteria.initialize_abstracts,
            Criteria.initialize_print_preview,
            Criteria.initialize_id,  # Must run after initialize_page_len().
        ],
    )


def create_feed_criteria(initial=None):
    return Criteria(
        initial=initial,
        options_initializers=[
            Criteria.initialize_page,
        ],
    )


class Criteria:
    """
    Represent a complete set of user-submitted search criteria.

    In practice, a criteria object is similar to `request.args`, but its
    internal structure uses multiple `MultiDict` objects instead of a single
    flat one like `request.args`:

    - `keywords` is a `MultiDict` of all keyword search-related parameters.
    - `filters` is a `MultiDict` of all filtering-related parameters.
    - `options` is a `MultiDict` of the remaining search parameters.

    The goal of this class is to provide a convenient place for storing and
    accessing validated parameters that describe a search request.
    """

    def __init__(self, initial=None, options_initializers=None):
        """
        Initialize the criteria.

        :param MultiDict|Criteria initial: Values to initialize from, either a
            `MultiDict` from `request.args`, or an instance of `Criteria`). If
            initial values are provided, they get deep-copied into the new
            `Criteria` instance.

        :param list options_initializers: Mandatory list of callables for
            validating and initializing the options from the `initial` values.
            The first argument to each callable is the `Criteria` instance to
            update with clean values, and the second one is the 'initial'
            `MultiDict` from which values can be extracted. An initializer *must
            not* assign a default value when an option has no 'initial' value,
            unless having a value is mandatory, otherwise all search URLs
            derived from the `Criteria` instance will end up with that value
            explicitly set.
        """
        self.keywords = MultiDict()
        self.filters = MultiDict()
        self.options = MultiDict()
        if initial:
            if isinstance(initial, Criteria):
                self.keywords = initial.keywords.deepcopy()
                self.filters = initial.filters.deepcopy()
                for initializer in options_initializers:
                    initializer(self, initial.options)
            else:
                assert isinstance(initial, MultiDict)
                self.initialize_keywords(initial)
                self.initialize_filters(initial)
                for initializer in options_initializers:
                    initializer(self, initial)

    def has_keywords(self):
        """Return `True` if criteria includes active keywords."""
        return any(self.keywords.values())

    def has_filters(self):
        """Return `True` if criteria includes active filters."""
        return any(self.filters.values())

    def is_searching(self):
        return self.has_keywords() or self.has_filters()

    def params(self, *, keywords=None, filters=None, options=None):
        """
        Return parameters for a search request.

        The returned parameters are based on the active criteria, but some may
        be overridden with the given values.

        :param MultiDict keywords: If not `None`, the argument overrides the
            active keywords in the returned values.

        :param MultiDict filters: If not `None`, the argument overrides the
            active filters in the returned values.

        :param dict options: A mapping whose values, if any, are to override the
            corresponding active options in the returned values.

        :return dict: A dict representing query string arguments in a search
            request (may be passed as the `values` argument to
            `flask.url_for()`).
        """
        query_params = MultiDict()
        query_params.update(self.keywords if keywords is None else keywords)
        query_params.update(self.filters if filters is None else filters)
        if options:
            # Update from those self.options that are not being overridden.
            for key in self.options.keys():
                if key not in options:
                    query_params.setlist(key, self.options.getlist(key))
            # Update with the overrides and additions.
            query_params.update(options)
        else:
            query_params.update(self.options)
        return query_params.to_dict(flat=False)

    def fit_page(self, page_count):
        """Ensure that the page number is less than or equal to the given page count."""
        self.options["page"] = min(self.options.get("page", 1), page_count)

    def get_active_sort_spec(self):
        """Return the spec of the active sort or, if none is active, the default sort to use."""
        sort_key = self.options.get("sort")
        if sort_key:
            return composer().sorts[sort_key]
        for sort_spec in composer().get_ordered_specs("sorts"):
            if sort_spec.is_allowed(self):
                return sort_spec
        return None

    def initialize_keywords(self, initial):
        for scope in composer().scopes.values():
            values = initial.getlist(scope.key)
            if values:
                self.keywords.setlist(scope.key, values)

    def initialize_filters(self, initial):
        for spec in composer().facets.values():
            values = initial.getlist(spec.filter_key)
            if values:
                self.filters.setlist(spec.filter_key, values)

    def initialize_page(self, initial):
        try:
            page = int(initial.get("page", 0))
            if page >= 1:
                self.options["page"] = page
        except ValueError:
            pass

    def initialize_page_len(self, initial):
        page_len = initial.get("page-len", 0)
        try:
            if page_len == "all":
                self.options["page-len"] = page_len
            elif int(page_len) >= 1:
                self.options["page-len"] = int(page_len)
        except ValueError:
            pass

    def initialize_sort(self, initial):
        sort_spec = composer().sorts.get(initial.get("sort", ""))
        if sort_spec and sort_spec.is_allowed(self):
            self.options["sort"] = sort_spec.key

    def initialize_abstracts(self, initial):
        if config("kerko.features.results_abstracts_toggler"):
            enabled_by_default = config("kerko.features.results_abstracts")
            abstracts = initial.get("abstracts")
            if abstracts:
                if abstracts in ["t", "1"] and not enabled_by_default:
                    self.options["abstracts"] = 1
                elif abstracts in ["f", "0"] and enabled_by_default:
                    self.options["abstracts"] = 0

    def initialize_print_preview(self, initial):
        if config("kerko.features.print_results") and initial.get("print-preview") in ["t", "1"]:
            self.options["print-preview"] = 1

    def initialize_id(self, initial):
        if self.options.get("page-len") == 1 and initial.get("id"):
            self.options["id"] = initial.get("id")
